#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from copy import deepcopy
from functools import wraps
from logging import Logger
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar

import pandas as pd
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generation_strategy_interface import GenerationStrategyInterface
from ax.core.generator_run import GeneratorRun
from ax.core.observation import ObservationFeatures
from ax.core.utils import (
    extend_pending_observations,
    get_pending_observation_features_based_on_trial_status,
)
from ax.exceptions.core import DataRequiredError, UnsupportedError, UserInputError
from ax.exceptions.generation_strategy import (
    GenerationStrategyCompleted,
    GenerationStrategyMisconfiguredException,
)
from ax.modelbridge.base import ModelBridge
from ax.modelbridge.generation_node import GenerationNode, GenerationStep
from ax.modelbridge.model_spec import FactoryFunctionModelSpec
from ax.modelbridge.modelbridge_utils import get_fixed_features_from_experiment
from ax.modelbridge.registry import _extract_model_state_after_gen, ModelRegistryBase
from ax.modelbridge.transition_criterion import TrialBasedCriterion
from ax.utils.common.logger import _round_floats_for_logging, get_logger
from ax.utils.common.typeutils import checked_cast, checked_cast_list, not_none
from pyre_extensions import none_throws

logger: Logger = get_logger(__name__)


MAX_CONDITIONS_GENERATED = 10000
MAX_GEN_DRAWS = 5
MAX_GEN_DRAWS_EXCEEDED_MESSAGE = (
    f"GenerationStrategy exceeded `MAX_GEN_DRAWS` of {MAX_GEN_DRAWS} while trying to "
    "generate a unique parameterization. This indicates that the search space has "
    "likely been fully explored, or that the sweep has converged."
)
T = TypeVar("T")


def step_based_gs_only(f: Callable[..., T]) -> Callable[..., T]:
    """
    For use as a decorator on functions only implemented for ``GenerationStep``-based
    ``GenerationStrategies``. Mainly useful for older ``GenerationStrategies``.
    """

    @wraps(f)
    def impl(
        self: "GenerationStrategy", *args: List[Any], **kwargs: Dict[str, Any]
    ) -> T:
        if self.is_node_based:
            raise UnsupportedError(
                f"{f.__name__} is not supported for GenerationNode based"
                " GenerationStrategies."
            )
        return f(self, *args, **kwargs)

    return impl


class GenerationStrategy(GenerationStrategyInterface):
    """GenerationStrategy describes which model should be used to generate new
    points for which trials, enabling and automating use of different models
    throughout the optimization process. For instance, it allows to use one
    model for the initialization trials, and another one for all subsequent
    trials. In the general case, this allows to automate use of an arbitrary
    number of models to generate an arbitrary numbers of trials
    described in the `trials_per_model` argument.

    Args:
        nodes: A list of `GenerationNode`. Each `GenerationNode` in the list
            represents a single node in a `GenerationStrategy` which, when
            composed of `GenerationNodes`, can be conceptualized as a graph instead
            of a linear list. `TransitionCriterion` defined in each `GenerationNode`
            represent the edges in the `GenerationStrategy` graph. `GenerationNodes`
            are more flexible than `GenerationSteps` and new `GenerationStrategies`
            should use nodes. Notably, either, but not both, of `nodes` and `steps`
            must be provided.
        steps: A list of `GenerationStep` describing steps of this strategy.
        name: An optional name for this generation strategy. If not specified,
            strategy's name will be names of its nodes' models joined with '+'.
    """

    _nodes: List[GenerationNode]
    _curr: GenerationNode  # Current node in the strategy.
    # Whether all models in this GS are in Models registry enum.
    _uses_registered_models: bool
    # All generator runs created through this generation strategy, in chronological
    # order.
    _generator_runs: List[GeneratorRun]
    # Experiment, for which this generation strategy has generated trials, if
    # it exists.
    _experiment: Optional[Experiment] = None
    _model: Optional[ModelBridge] = None  # Current model.

    def __init__(
        self,
        steps: Optional[List[GenerationStep]] = None,
        name: Optional[str] = None,
        nodes: Optional[List[GenerationNode]] = None,
    ) -> None:
        # Validate that one and only one of steps or nodes is provided
        if not ((steps is None) ^ (nodes is None)):
            raise GenerationStrategyMisconfiguredException(
                error_info="GenerationStrategy must contain either steps or nodes."
            )

        # pyre-ignore[8]
        self._nodes = none_throws(nodes if steps is None else steps)

        # Validate correctness of steps list or nodes graph
        if isinstance(steps, list) and all(
            isinstance(s, GenerationStep) for s in steps
        ):
            self._validate_and_set_step_sequence(steps=self._nodes)
        elif isinstance(nodes, list) and self.is_node_based:
            self._validate_and_set_node_graph(nodes=nodes)
        else:
            # TODO[mgarrard]: Allow mix of nodes and steps
            raise GenerationStrategyMisconfiguredException(
                "`GenerationStrategy` inputs are:\n"
                "`steps` (list of `GenerationStep`) or\n"
                "`nodes` (list of `GenerationNode`)."
            )

        # Log warning if the GS uses a non-registered (factory function) model.
        self._uses_registered_models = not any(
            isinstance(ms, FactoryFunctionModelSpec)
            for node in self._nodes
            for ms in node.model_specs
        )
        if not self._uses_registered_models:
            logger.info(
                "Using model via callable function, "
                "so optimization is not resumable if interrupted."
            )
        self._generator_runs = []
        # Set name to an explicit value ahead of time to avoid
        # adding properties during equality checks
        super().__init__(name=name or self._make_default_name())

    @property
    def is_node_based(self) -> bool:
        """Whether this strategy consists of GenerationNodes only.
        This is useful for determining initialization properties and
        other logic.
        """
        return not any(isinstance(n, GenerationStep) for n in self._nodes) and all(
            isinstance(n, GenerationNode) for n in self._nodes
        )

    @property
    def name(self) -> str:
        """Name of this generation strategy. Defaults to a combination of model
        names provided in generation steps, set at the time of the
        ``GenerationStrategy`` creation.
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """Set generation strategy name."""
        self._name = name

    @property
    @step_based_gs_only
    def model_transitions(self) -> List[int]:
        """[DEPRECATED]List of trial indices where a transition happened from one model
        to another.
        """
        raise DeprecationWarning(
            "`model_transitions` is no longer supported. Please refer to `model_key` "
            "field on generator runs for similar information if needed."
        )

    @property
    def current_step(self) -> GenerationStep:
        """Current generation step."""
        if not isinstance(self._curr, GenerationStep):
            raise TypeError(
                "The current object is not a GenerationStep, you may be looking "
                "for the current_node property."
            )
        return self._curr

    @property
    def current_node(self) -> GenerationNode:
        """Current generation node."""
        if not isinstance(self._curr, GenerationNode):
            raise TypeError(
                "The current object is not a GenerationNode, you may be looking for the"
                " current_step property."
            )
        return self._curr

    @property
    def current_node_name(self) -> str:
        """Current generation node name."""
        return self._curr.node_name

    @property
    @step_based_gs_only
    def current_step_index(self) -> int:
        """Returns the index of the current generation step. This attribute
        is replaced by node_name in newer GenerationStrategies but surfaced here
        for backward compatibility.
        """
        node_names_for_all_steps = [step._node_name for step in self._nodes]
        assert (
            self._curr.node_name in node_names_for_all_steps
        ), "The current step is not found in the list of steps"

        return node_names_for_all_steps.index(self._curr.node_name)

    @property
    def model(self) -> Optional[ModelBridge]:
        """Current model in this strategy. Returns None if no model has been set
        yet (i.e., if no generator runs have been produced from this GS).
        """
        return self._curr._fitted_model

    @property
    def experiment(self) -> Experiment:
        """Experiment, currently set on this generation strategy."""
        if self._experiment is None:
            raise ValueError("No experiment set on generation strategy.")
        return not_none(self._experiment)

    @experiment.setter
    def experiment(self, experiment: Experiment) -> None:
        """If there is an experiment set on this generation strategy as the
        experiment it has been generating generator runs for, check if the
        experiment passed in is the same as the one saved and log an information
        statement if its not. Set the new experiment on this generation strategy.
        """
        if self._experiment is None or experiment._name == self.experiment._name:
            self._experiment = experiment
        else:
            raise ValueError(
                "This generation strategy has been used for experiment "
                f"{self.experiment._name} so far; cannot reset experiment"
                f" to {experiment._name}. If this is a new optimization, "
                "a new generation strategy should be created instead."
            )

    @property
    def last_generator_run(self) -> Optional[GeneratorRun]:
        """Latest generator run produced by this generation strategy.
        Returns None if no generator runs have been produced yet.
        """
        # Used to restore current model when decoding a serialized GS.
        return self._generator_runs[-1] if self._generator_runs else None

    @property
    def uses_non_registered_models(self) -> bool:
        """Whether this generation strategy involves models that are not
        registered and therefore cannot be stored."""
        return not self._uses_registered_models

    @property
    def trials_as_df(self) -> Optional[pd.DataFrame]:
        """Puts information on individual trials into a data frame for easy
        viewing.

        For example for a GenerationStrategy composed of GenerationSteps:
        Gen. Step | Models   | Trial Index | Trial Status | Arm Parameterizations
        [0]       | [Sobol]  | 0           | RUNNING      | {"0_0":{"x":9.17...}}
        """
        logger.info(
            "Note that parameter values in dataframe are rounded to 2 decimal "
            "points; the values in the dataframe are thus not the exact ones "
            "suggested by Ax in trials."
        )
        if self._experiment is None or len(self.experiment.trials) == 0:
            return None

        step_or_node_col = (
            "Generation Nodes" if self.is_node_based else "Generation Step"
        )
        records = [
            {
                step_or_node_col: [
                    (
                        gr._generation_node_name
                        if gr.generator_run_type != "MANUAL"
                        else "MANUAL"
                    )
                    for gr in trial.generator_runs
                ],
                "Generation Model(s)": [
                    (gr._model_key if gr.generator_run_type != "MANUAL" else "MANUAL")
                    for gr in trial.generator_runs
                ],
                "Trial Index": trial_idx,
                "Trial Status": trial.status.name,
                "Arm Parameterizations": {
                    arm.name: _round_floats_for_logging(arm.parameters)
                    for arm in trial.arms
                },
            }
            for trial_idx, trial in self.experiment.trials.items()
        ]

        return pd.DataFrame.from_records(records).reindex(
            columns=[
                step_or_node_col,
                "Generation Model(s)",
                "Trial Index",
                "Trial Status",
                "Arm Parameterizations",
            ]
        )

    @property
    def optimization_complete(self) -> bool:
        """Checks whether all nodes are completed in the generation strategy."""
        return all(node.is_completed for node in self._nodes)

    @property
    @step_based_gs_only
    def _steps(self) -> List[GenerationStep]:
        """List of generation steps."""
        return self._nodes  # pyre-ignore[7]

    def gen(
        self,
        experiment: Experiment,
        data: Optional[Data] = None,
        n: int = 1,
        pending_observations: Optional[Dict[str, List[ObservationFeatures]]] = None,
        **kwargs: Any,
    ) -> GeneratorRun:
        """Produce the next points in the experiment. Additional kwargs passed to
        this method are propagated directly to the underlying model's `gen`, along
        with the `model_gen_kwargs` set on the current generation node.

        NOTE: Each generator run returned from this function must become a single
        trial on the experiment to comply with assumptions made in generation
        strategy. Do not split one generator run produced from generation strategy
        into multiple trials (never making a generator run into a trial is allowed).

        Args:
            experiment: Experiment, for which the generation strategy is producing
                a new generator run in the course of `gen`, and to which that
                generator run will be added as trial(s). Information stored on the
                experiment (e.g., trial statuses) is used to determine which model
                will be used to produce the generator run returned from this method.
            data: Optional data to be passed to the underlying model's `gen`, which
                is called within this method and actually produces the resulting
                generator run. By default, data is all data on the `experiment`.
            n: Integer representing how many arms should be in the generator run
                produced by this method. NOTE: Some underlying models may ignore
                the `n` and produce a model-determined number of arms. In that
                case this method will also output a generator run with number of
                arms that can differ from `n`.
            pending_observations: A map from metric name to pending
                observations for that metric, used by some models to avoid
                resuggesting points that are currently being evaluated.
        """
        return self._gen_multiple(
            experiment=experiment,
            num_generator_runs=1,
            data=data,
            n=n,
            pending_observations=pending_observations,
            **kwargs,
        )[0]

    def gen_with_multiple_nodes(
        self,
        experiment: Experiment,
        data: Optional[Data] = None,
        pending_observations: Optional[Dict[str, List[ObservationFeatures]]] = None,
        arms_per_node: Optional[Dict[str, int]] = None,
    ) -> List[GeneratorRun]:
        """Produces a List of GeneratorRuns for a single trial, either ``Trial`` or
        ``BatchTrial``, and if producing a ``BatchTrial`` allows for multiple
        models to be used to generate GeneratorRuns for that trial.

        NOTE: This method is in development.  Please do not use it yet.

        Args:
            experiment: Experiment, for which the generation strategy is producing
                a new generator run in the course of `gen`, and to which that
                generator run will be added as trial(s). Information stored on the
                experiment (e.g., trial statuses) is used to determine which model
                will be used to produce the generator run returned from this method.
            data: Optional data to be passed to the underlying model's `gen`, which
                is called within this method and actually produces the resulting
                generator run. By default, data is all data on the `experiment`.
            pending_observations: A map from metric name to pending
                observations for that metric, used by some models to avoid
                resuggesting points that are currently being evaluated.
            arms_per_node: A map from node name to the number of arms to generate
                from that node. If not provided, the number of arms to generate
                from each node defaults to one.

        Returns:
            A list of ``GeneratorRuns`` for a single trial.
        """
        # TODO: @mgarrard merge into gen method, just starting here to derisk
        # Validate `arms_per_node` if specified, otherwise construct the default
        # behavior with keys being node names and values being 1 to represent
        # generating a single GR from each node.
        node_names = [node.node_name for node in self._nodes]
        if arms_per_node is not None and not all(
            node_name in arms_per_node for node_name in node_names
        ):
            raise UserInputError(
                f"""
                Each node defined in the GenerationStrategy must have
                an associated number of arms to generate from that node
                defined in `arms_per_node`. {arms_per_node} does not
                include all of {node_names}. It may be helpful to double check
                the spelling.
                """
            )
        if arms_per_node is None:
            arms_per_node = {node_name: 1 for node_name in node_names}
        grs = []
        continue_gen_for_trial = True

        while continue_gen_for_trial:
            next_node_name = self.current_node_name
            should_transition, next_node = self._curr.should_transition_to_next_node(
                raise_data_required_error=False
            )
            if should_transition:
                assert next_node is not None
                next_node_name = next_node
            grs.extend(
                self._gen_multiple(
                    experiment=experiment,
                    num_generator_runs=1,
                    data=data,
                    n=arms_per_node[next_node_name],
                    pending_observations=pending_observations,
                )
            )
            continue_gen_for_trial = self._should_continue_gen_for_trial()
        return grs

    def gen_for_multiple_trials_with_multiple_models(
        self,
        experiment: Experiment,
        num_generator_runs: int,
        data: Optional[Data] = None,
        n: int = 1,
    ) -> List[List[GeneratorRun]]:
        """Produce GeneratorRuns for multiple trials at once with the possibility of
        ensembling, or using multiple models per trial, getting multiple
        GeneratorRuns per trial.

        NOTE: This method is in development.  Please do not use it yet.

        Args:
            experiment: Experiment, for which the generation strategy is producing
                a new generator run in the course of `gen`, and to which that
                generator run will be added as trial(s). Information stored on the
                experiment (e.g., trial statuses) is used to determine which model
                will be used to produce the generator run returned from this method.
            data: Optional data to be passed to the underlying model's `gen`, which
                is called within this method and actually produces the resulting
                generator run. By default, data is all data on the `experiment`.
            n: Integer representing how many trials should be in the generator run
                produced by this method. NOTE: Some underlying models may ignore
                the ``n`` and produce a model-determined number of arms. In that
                case this method will also output a generator run with number of
                arms that can differ from ``n``.

        Returns:
            A list of lists of lists generator runs. Each outer list represents
            a trial being suggested and  each inner list represents a generator
            run for that trial.
        """
        grs = self._gen_multiple(
            experiment=experiment,
            num_generator_runs=num_generator_runs,
            data=data,
            n=n,
            pending_observations=get_pending_observation_features_based_on_trial_status(
                experiment=experiment
            ),
            fixed_features=get_fixed_features_from_experiment(experiment=experiment),
        )
        return [[gr] for gr in grs]

    def current_generator_run_limit(
        self,
    ) -> Tuple[int, bool]:
        """First check if we can move the generation strategy to the next node, which
        is safe, as the next call to ``gen`` will just pick up from there. Then
        determine how many generator runs this generation strategy can generate right
        now, assuming each one of them becomes its own trial, and whether optimization
        is completed.

        Returns: a two-item tuple of:
              - the number of generator runs that can currently be produced, with -1
                meaning unlimited generator runs,
              - whether optimization is completed and the generation strategy cannot
                generate any more generator runs at all.
        """
        try:
            self._maybe_transition_to_next_node(raise_data_required_error=False)
        except GenerationStrategyCompleted:
            return 0, True

        # if the generation strategy is not complete, optimization is not complete
        return self._curr.generator_run_limit(), False

    def clone_reset(self) -> GenerationStrategy:
        """Copy this generation strategy without it's state."""
        cloned_nodes = deepcopy(self._nodes)
        for n in cloned_nodes:
            # Unset the generation strategy back-pointer, so the nodes are not
            # associated with any generation strategy.
            n._generation_strategy = None
        if self.is_node_based:
            return GenerationStrategy(name=self.name, nodes=cloned_nodes)

        return GenerationStrategy(
            name=self.name, steps=checked_cast_list(GenerationStep, cloned_nodes)
        )

    def _unset_non_persistent_state_fields(self) -> None:
        """Utility for testing convenience: unset fields of generation strategy
        that are set during candidate generation; these fields are not persisted
        during storage. To compare a pre-storage and a reloaded generation
        strategies; call this utility on the pre-storage one first. The rest
        of the fields should be identical.
        """
        self._model = None
        for s in self._nodes:
            s._model_spec_to_gen_from = None

    @step_based_gs_only
    def _validate_and_set_step_sequence(self, steps: List[GenerationStep]) -> None:
        """Initialize and validate the steps provided to this GenerationStrategy.

        Some GenerationStrategies are composed of GenerationStep objects, but we also
        need to initialize the correct GenerationNode representation for these steps.
        This function validates:
            1. That only the last step has num_trials=-1, which indicates unlimited
               trial generation is possible.
            2. That each step's num_trials attribute is either positive or -1
            3. That each step's max_parallelism attribute is either None or positive
        It then sets the correct TransitionCriterion and node_name attributes on the
        underlying GenerationNode objects.
        """
        for idx, step in enumerate(steps):
            if step.num_trials == -1 and len(step.completion_criteria) < 1:
                if idx < len(self._steps) - 1:
                    raise UserInputError(
                        "Only last step in generation strategy can have "
                        "`num_trials` set to -1 to indicate that the model in "
                        "the step should be used to generate new trials "
                        "indefinitely unless completion criteria present."
                    )
            elif step.num_trials < 1 and step.num_trials != -1:
                raise UserInputError(
                    "`num_trials` must be positive or -1 (indicating unlimited) "
                    "for all generation steps."
                )
            if step.max_parallelism is not None and step.max_parallelism < 1:
                raise UserInputError(
                    "Maximum parallelism should be None (if no limit) or "
                    f"a positive number. Got: {step.max_parallelism} for "
                    f"step {step.model_name}."
                )

            step._node_name = f"GenerationStep_{str(idx)}"
            step.index = idx

            # Set transition_to field for all but the last step, which remains
            # null.
            if idx != len(self._steps):
                for transition_criteria in step.transition_criteria:
                    if (
                        transition_criteria.criterion_class
                        != "MaxGenerationParallelism"
                    ):
                        transition_criteria._transition_to = (
                            f"GenerationStep_{str(idx + 1)}"
                        )
            step._generation_strategy = self
        self._curr = steps[0]

    def _validate_and_set_node_graph(self, nodes: List[GenerationNode]) -> None:
        """Initialize and validate the node graph provided to this GenerationStrategy.

        This function validates:
            1. That all nodes have unique names.
            2. That there is at least one node with a transition_to field.
            3. That all `transition_to` attributes on a TransitionCriterion point to
                another node in the same GenerationStrategy.
            4. Warns if no nodes contain a transition criterion
        """
        node_names = []
        for node in self._nodes:
            # validate that all node names are unique
            if node.node_name in node_names:
                raise GenerationStrategyMisconfiguredException(
                    error_info="All node names in a GenerationStrategy "
                    + "must be unique."
                )

            node_names.append(node.node_name)
            node._generation_strategy = self

        # Validate that the next_node is in the ``GenerationStrategy`` and that all
        # TCs in one "transition edge" (so all TCs from one node to another) have the
        # same `continue_trial_generation` setting. Since multiple TCs together
        # constitute one "transition edge", not having all TCs on such an "edge"
        # indicate the same resulting state (continuing generation for same trial
        # vs. stopping it after generating from current node) would indicate a
        # malformed generation node DAG definition and therefore a
        # malformed ``GenerationStrategy``.
        contains_a_transition_to_argument = False
        for node in self._nodes:
            for next_node, tcs in node.transition_edges.items():
                contains_a_transition_to_argument = True
                if next_node is None:
                    # TODO: @mgarrard remove MaxGenerationParallelism check when
                    # we update TransitionCriterion always define `transition_to`
                    for tc in tcs:
                        if "MaxGenerationParallelism" not in tc.criterion_class:
                            raise GenerationStrategyMisconfiguredException(
                                error_info="Only MaxGenerationParallelism transition"
                                " criterion can have a null `transition_to` argument,"
                                f" but {tc.criterion_class} does not define "
                                f"`transition_to` on {node.node_name}."
                            )
                if next_node is not None and next_node not in node_names:
                    raise GenerationStrategyMisconfiguredException(
                        error_info=f"`transition_to` argument "
                        f"{next_node} does not correspond to any node in"
                        " this GenerationStrategy."
                    )
                if (
                    next_node is not None
                    and len({tc.continue_trial_generation for tc in tcs}) > 1
                ):
                    raise GenerationStrategyMisconfiguredException(
                        error_info=f"All transition criteria on an edge "
                        f"from node {node.node_name} to node {next_node} "
                        "should have the same `continue_trial_generation` "
                        "setting."
                    )

        # validate that at least one node has transition_to field
        if len(self._nodes) > 1 and not contains_a_transition_to_argument:
            logger.warning(
                "None of the nodes in this GenerationStrategy "
                "contain a `transition_to` argument in their transition_criteria. "
                "Therefore, the GenerationStrategy will not be able to "
                "move from one node to another. Please add a "
                "`transition_to` argument."
            )
        self._curr = nodes[0]

    @step_based_gs_only
    def _step_repr(self, step_str_rep: str) -> str:
        """Return the string representation of the steps in a GenerationStrategy
        composed of GenerationSteps.
        """
        step_str_rep += "steps=["
        remaining_trials = "subsequent" if len(self._nodes) > 1 else "all"
        for step in self._nodes:
            num_trials = remaining_trials
            for criterion in step.transition_criteria:
                if criterion.criterion_class == "MaxTrials" and isinstance(
                    criterion, TrialBasedCriterion
                ):
                    num_trials = criterion.threshold

            try:
                model_name = step.model_spec_to_gen_from.model_key
            except TypeError:
                model_name = "model with unknown name"

            step_str_rep += f"{model_name} for {num_trials} trials, "
        step_str_rep = step_str_rep[:-2]
        step_str_rep += "])"
        return step_str_rep

    def _make_default_name(self) -> str:
        """Make a default name for this generation strategy; used when no name is passed
        to the constructor. For node-based generation strategies, the name is
        constructed by joining together the names of the nodes set on this
        generation strategy. For step-based generation strategies, the model keys
        of the underlying model specs are used.
        Note: This should only be called once the nodes are set.
        """
        if not self._nodes:
            raise UnsupportedError(
                "Cannot make a default name for a generation strategy with no nodes "
                "set yet."
            )
        # TODO: Simplify this after updating GStep names to represent underlying models.
        if self.is_node_based:
            node_names = (node.node_name for node in self._nodes)
        else:
            node_names = (node.model_spec_to_gen_from.model_key for node in self._nodes)
            # Trim the "get_" beginning of the factory function if it's there.
            node_names = (n[4:] if n[:4] == "get_" else n for n in node_names)
        return "+".join(node_names)

    def __repr__(self) -> str:
        """String representation of this generation strategy."""
        gs_str = f"GenerationStrategy(name='{self.name}', "
        if not self.is_node_based:
            return self._step_repr(gs_str)
        gs_str += f"nodes={str(self._nodes)})"
        return gs_str

    # ------------------------- Candidate generation helpers. -------------------------

    def _gen_multiple(
        self,
        experiment: Experiment,
        num_generator_runs: int,
        data: Optional[Data] = None,
        n: int = 1,
        pending_observations: Optional[Dict[str, List[ObservationFeatures]]] = None,
        **model_gen_kwargs: Any,
    ) -> List[GeneratorRun]:
        """Produce multiple generator runs at once, to be made into multiple
        trials on the experiment.

        NOTE: This is used to ensure that maximum parallelism and number
        of trials per node are not violated when producing many generator
        runs from this generation strategy in a row. Without this function,
        if one generates multiple generator runs without first making any
        of them into running trials, generation strategy cannot enforce that it only
        produces as many generator runs as are allowed by the parallelism
        limit and the limit on number of trials in current node.

        Args:
            experiment: Experiment, for which the generation strategy is producing
                a new generator run in the course of `gen`, and to which that
                generator run will be added as trial(s). Information stored on the
                experiment (e.g., trial statuses) is used to determine which model
                will be used to produce the generator run returned from this method.
            data: Optional data to be passed to the underlying model's `gen`, which
                is called within this method and actually produces the resulting
                generator run. By default, data is all data on the `experiment`.
            n: Integer representing how many arms should be in the generator run
                produced by this method. NOTE: Some underlying models may ignore
                the ``n`` and produce a model-determined number of arms. In that
                case this method will also output a generator run with number of
                arms that can differ from ``n``.
            pending_observations: A map from metric name to pending
                observations for that metric, used by some models to avoid
                resuggesting points that are currently being evaluated.
            model_gen_kwargs: Keyword arguments that are passed through to
                ``GenerationNode.gen``, which will pass them through to
                ``ModelSpec.gen``, which will pass them to ``ModelBridge.gen``.
        """
        self.experiment = experiment
        self._maybe_transition_to_next_node()
        self._fit_current_model(data=data)

        # Get GeneratorRun limit that respects the node's transition criterion that
        # affect the number of generator runs that can be produced.
        gr_limit = self._curr.generator_run_limit(raise_generation_errors=True)
        if gr_limit == -1:
            num_generator_runs = max(num_generator_runs, 1)
        else:
            num_generator_runs = max(min(num_generator_runs, gr_limit), 1)
        generator_runs = []
        pending_observations = deepcopy(pending_observations) or {}
        for _ in range(num_generator_runs):
            try:
                generator_run = self._curr.gen(
                    n=n,
                    pending_observations=pending_observations,
                    arms_by_signature_for_deduplication=(
                        experiment.arms_by_signature_for_deduplication
                    ),
                    **model_gen_kwargs,
                )

            except DataRequiredError as err:
                # Model needs more data, so we log the error and return
                # as many generator runs as we were able to produce, unless
                # no trials were produced at all (in which case its safe to raise).
                if len(generator_runs) == 0:
                    raise
                logger.debug(f"Model required more data: {err}.")
                break

            self._generator_runs.append(generator_run)
            generator_runs.append(generator_run)

            # Extend the `pending_observation` with newly generated point(s)
            # to avoid repeating them.
            extend_pending_observations(
                experiment=experiment,
                pending_observations=pending_observations,
                generator_run=generator_run,
            )
        return generator_runs

    def _should_continue_gen_for_trial(self) -> bool:
        """Determine if we should continue generating for the current trial, or end
        generation for the current trial. Note that generating more would involve
        transitioning to a next node, because each node generates once per call to
        ``GenerationStrategy.gen_with_multiple_nodes``.

        Returns:
            A boolean which represents if generation for a trial is complete
        """
        should_transition, next_node = self._curr.should_transition_to_next_node(
            raise_data_required_error=False
        )
        # if we should not transition nodes, we should stop generation for this trial.
        if not should_transition:
            return False

        # if we will transition nodes, check if the transition criterion which define
        # the transition from this node to the next node indicate that we should
        # continue generating in the same trial, otherwise end the generation.
        assert next_node is not None
        return all(
            tc.continue_trial_generation
            for tc in self._curr.transition_edges[next_node]
        )

    # ------------------------- Model selection logic helpers. -------------------------

    def _fit_current_model(self, data: Optional[Data]) -> None:
        """Fits or update the model on the current generation node (does not move
        between generation nodes).

        Args:
            data: Optional ``Data`` to fit or update with; if not specified, generation
                strategy will obtain the data via ``experiment.lookup_data``.
        """
        data = self.experiment.lookup_data() if data is None else data
        # If last generator run's index matches the current node, extract
        # model state from last generator run and pass it to the model
        # being instantiated in this function.
        model_state_on_lgr = self._get_model_state_from_last_generator_run()
        if not data.df.empty:
            trial_indices_in_data = sorted(data.df["trial_index"].unique())
            logger.debug(f"Fitting model with data for trials: {trial_indices_in_data}")

        self._curr.fit(experiment=self.experiment, data=data, **model_state_on_lgr)
        self._model = self._curr._fitted_model

    def _maybe_transition_to_next_node(
        self, raise_data_required_error: bool = True
    ) -> bool:
        """Moves this generation strategy to next node if the current node is completed,
        and it is not the last node in this generation strategy. This method is safe to
        use both when generating candidates or simply checking how many generator runs
        (to be made into trials) can currently be produced.

        NOTE: this method raises ``GenerationStrategyCompleted`` error if the current
        generation node is complete, but it is also the last in generation strategy.

        Args:
            raise_data_required_error: Whether to raise ``DataRequiredError`` in the
                maybe_step_completed method in GenerationNode class.

        Returns:
            Whether generation strategy moved to the next node.
        """
        move_to_next_node, next_node = self._curr.should_transition_to_next_node(
            raise_data_required_error=raise_data_required_error
        )
        if move_to_next_node:
            if self.optimization_complete:
                raise GenerationStrategyCompleted(
                    f"Generation strategy {self} generated all the trials as "
                    "specified in its nodes."
                )
            if next_node is None:
                # If the last node did not specify which node to transition to,
                # move to the next node in the list.
                current_node_index = self._nodes.index(self._curr)
                next_node = self._nodes[current_node_index + 1].node_name
            for node in self._nodes:
                if node.node_name == next_node:
                    self._curr = node
                    # Moving to the next node also entails unsetting this GS's model
                    # (since new node's model will be initialized for the first time;
                    # this is done in `self._fit_current_model).
                    self._model = None
        return move_to_next_node

    def _get_model_state_from_last_generator_run(self) -> Dict[str, Any]:
        lgr = self.last_generator_run
        model_state_on_lgr = {}
        # Need to check if model_spec_to_gen_from is none to account for
        # ExternalGenerationNodes which leverage models from outside Ax.
        model_on_curr = (
            self._curr.model_spec_to_gen_from.model_enum
            if self._curr.model_spec_to_gen_from
            else None
        )
        if lgr is None:
            return model_state_on_lgr

        if not self.is_node_based:
            grs_equal = lgr._generation_step_index == self.current_step_index
        else:
            grs_equal = lgr._generation_node_name == self._curr.node_name

        if grs_equal and lgr._model_state_after_gen:
            if self.model or isinstance(model_on_curr, ModelRegistryBase):
                model_cls = (
                    self.model.model.__class__
                    if self.model is not None
                    # NOTE: This checked cast is save per the OR-statement in last line
                    # of the IF-check above.
                    else checked_cast(ModelRegistryBase, model_on_curr).model_class
                )
                model_state_on_lgr = _extract_model_state_after_gen(
                    generator_run=lgr,
                    model_class=model_cls,
                )

        return model_state_on_lgr



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any, Dict, Optional, Type, Union

from ax.benchmark.benchmark_method import (
    BenchmarkMethod,
    get_benchmark_scheduler_options,
)
from ax.modelbridge.generation_node import GenerationStep
from ax.modelbridge.generation_strategy import GenerationStrategy
from ax.modelbridge.registry import Models
from ax.models.torch.botorch_modular.model import SurrogateSpec
from ax.service.scheduler import SchedulerOptions
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.acquisition.analytic import LogExpectedImprovement
from botorch.acquisition.logei import qLogNoisyExpectedImprovement
from botorch.acquisition.multi_objective.logei import (
    qLogNoisyExpectedHypervolumeImprovement,
)
from botorch.acquisition.multi_objective.monte_carlo import (
    qNoisyExpectedHypervolumeImprovement,
)
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.model import Model


model_names_abbrevations: Dict[str, str] = {
    SaasFullyBayesianSingleTaskGP.__name__: "SAAS",
}
acqf_name_abbreviations: Dict[str, str] = {
    qLogNoisyExpectedImprovement.__name__: "qLogNEI",
    qNoisyExpectedHypervolumeImprovement.__name__: "qNEHVI",
    qLogNoisyExpectedHypervolumeImprovement.__name__: "qLogNEHVI",
    LogExpectedImprovement.__name__: "LogEI",
}


def get_sobol_botorch_modular_acquisition(
    model_cls: Type[Model],
    acquisition_cls: Type[AcquisitionFunction],
    distribute_replications: bool,
    scheduler_options: Optional[SchedulerOptions] = None,
    name: Optional[str] = None,
    num_sobol_trials: int = 5,
    model_gen_kwargs: Optional[Dict[str, Any]] = None,
) -> BenchmarkMethod:
    """Get a `BenchmarkMethod` that uses Sobol followed by MBM.

    Args:
        model_cls: BoTorch model class, e.g. SingleTaskGP
        acquisition_cls: Acquisition function class, e.g.
            `qLogNoisyExpectedImprovement`.
        distribute_replications: Whether to use multiple machines
        scheduler_options: Passed as-is to scheduler. Default:
            `get_benchmark_scheduler_options()`.
        name: Name that will be attached to the `GenerationStrategy`.
        num_sobol_trials: Number of Sobol trials; if the scheduler_options
            specify to use `BatchTrial`s, then this refers to the number of
            `BatchTrial`s.
        model_gen_kwargs: Passed to the BoTorch `GenerationStep` and ultimately
            to the BoTorch `Model`.

    Example:
        >>> # A simple example
        >>> from ax.benchmark.methods.sobol_botorch_modular import (
        ...     get_sobol_botorch_modular_acquisition
        ... )
        >>> from ax.benchmark.benchmark_method import get_benchmark_scheduler_options
        >>>
        >>> method = get_sobol_botorch_modular_acquisition(
        ...     model_cls=SingleTaskGP,
        ...     acquisition_cls=qLogNoisyExpectedImprovement,
        ...     distribute_replications=False,
        ... )
        >>> # Pass sequential=False to BoTorch's optimize_acqf
        >>> batch_method = get_sobol_botorch_modular_acquisition(
        ...     model_cls=SingleTaskGP,
        ...     acquisition_cls=qLogNoisyExpectedImprovement,
        ...     distribute_replications=False,
        ...     scheduler_options=get_benchmark_scheduler_options(
        ...         batch_size=5,
        ...     ),
        ...     model_gen_kwargs={
        ...         "model_gen_options": {
        ...             "optimizer_kwargs": {"sequential": False}
        ...         }
        ...     },
        ...     num_sobol_trials=1,
        ... )
    """
    model_kwargs: Dict[
        str, Union[Type[AcquisitionFunction], Dict[str, SurrogateSpec], bool]
    ] = {
        "botorch_acqf_class": acquisition_cls,
        "surrogate_specs": {"BoTorch": SurrogateSpec(botorch_model_class=model_cls)},
    }

    model_name = model_names_abbrevations.get(model_cls.__name__, model_cls.__name__)
    acqf_name = acqf_name_abbreviations.get(
        acquisition_cls.__name__, acquisition_cls.__name__
    )
    # Historically all benchmarks were sequential, so sequential benchmarks
    # don't get anything added to their name, for continuity
    batch_suffix = ""
    if (
        scheduler_options is not None
        and (batch_size := scheduler_options.batch_size) is not None
    ):
        if batch_size > 1:
            batch_suffix = f"_q{batch_size}"

    name = name or f"MBM::{model_name}_{acqf_name}{batch_suffix}"

    generation_strategy = GenerationStrategy(
        name=name,
        steps=[
            GenerationStep(
                model=Models.SOBOL,
                num_trials=num_sobol_trials,
                min_trials_observed=num_sobol_trials,
            ),
            GenerationStep(
                model=Models.BOTORCH_MODULAR,
                num_trials=-1,
                model_kwargs=model_kwargs,
                model_gen_kwargs=model_gen_kwargs or {},
            ),
        ],
    )

    return BenchmarkMethod(
        name=generation_strategy.name,
        generation_strategy=generation_strategy,
        scheduler_options=scheduler_options or get_benchmark_scheduler_options(),
        distribute_replications=distribute_replications,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from itertools import product

import torch
from ax.exceptions.core import AxError
from ax.models.torch.botorch_modular.kernels import ScaleMaternKernel, TemporalKernel
from ax.utils.common.testutils import TestCase
from gpytorch.constraints import Positive
from gpytorch.kernels import MaternKernel, PeriodicKernel
from gpytorch.priors import GammaPrior


class KernelsTest(TestCase):
    def test_scalematern_kernel(self) -> None:
        covar = ScaleMaternKernel(
            ard_num_dims=10,
            lengthscale_prior=GammaPrior(6.0, 3.0),
            outputscale_prior=GammaPrior(2.0, 0.15),
            batch_shape=torch.Size([2]),
        )
        self.assertTrue(isinstance(covar.base_kernel, MaternKernel))
        self.assertTrue(isinstance(covar.base_kernel, MaternKernel))
        self.assertEqual(covar.base_kernel.ard_num_dims, 10)
        self.assertEqual(covar.base_kernel.lengthscale_prior.rate, 3.0)
        self.assertEqual(covar.base_kernel.lengthscale_prior.concentration, 6.0)
        self.assertEqual(covar.outputscale_prior.rate, 0.15)
        self.assertEqual(covar.outputscale_prior.concentration, 2.0)
        self.assertEqual(covar.base_kernel.batch_shape[0], 2)

    def test_temporal_kernel(self) -> None:
        ls_prior = GammaPrior(6.0, 3.0)
        os_prior = GammaPrior(2.0, 0.15)
        tls_prior = GammaPrior(3.0, 6.0)
        pl_prior = GammaPrior(2.0, 6.0)
        temporal_features = [8, 9]
        ls_constraint = Positive()
        os_constraint = Positive()
        tls_constraint = Positive()
        pl_constraint = Positive()
        for fixed_period_length, batch_shape, matern_ard_num_dims in product(
            (6.0, None), (torch.Size([]), torch.Size([2])), (10, None)
        ):
            covar = TemporalKernel(
                dim=12,
                matern_ard_num_dims=matern_ard_num_dims,
                temporal_features=temporal_features,
                batch_shape=batch_shape,
                lengthscale_prior=ls_prior,
                outputscale_prior=os_prior,
                temporal_lengthscale_prior=tls_prior,
                period_length_prior=pl_prior if fixed_period_length is None else None,
                fixed_period_length=fixed_period_length,
                lengthscale_constraint=ls_constraint,
                outputscale_constraint=os_constraint,
                period_length_constraint=(
                    pl_constraint if fixed_period_length is None else None
                ),
                temporal_lengthscale_constraint=tls_constraint,
            )
            self.assertTrue(isinstance(covar.base_kernel.kernels[0], MaternKernel))
            self.assertTrue(isinstance(covar.base_kernel.kernels[1], PeriodicKernel))
            matern, periodic = covar.base_kernel.kernels

            self.assertEqual(matern.ard_num_dims, matern_ard_num_dims)
            self.assertEqual(
                matern.active_dims.tolist(),
                list(set(range(12)) - set(temporal_features)),
            )
            self.assertIs(matern.lengthscale_prior, ls_prior)
            self.assertEqual(periodic.ard_num_dims, 2)
            self.assertEqual(periodic.active_dims.tolist(), temporal_features)
            self.assertIs(periodic.lengthscale_prior, tls_prior)
            if fixed_period_length is None:
                self.assertIs(periodic.period_length_prior, pl_prior)
                self.assertIs(periodic.raw_period_length_constraint, pl_constraint)
            self.assertIs(matern.raw_lengthscale_constraint, ls_constraint)
            self.assertIs(periodic.raw_lengthscale_constraint, tls_constraint)
            self.assertIs(covar.raw_outputscale_constraint, os_constraint)
            self.assertIs(covar.outputscale_prior, os_prior)
            self.assertEqual(matern.batch_shape, batch_shape)
            self.assertEqual(periodic.batch_shape, batch_shape)
            if fixed_period_length is not None:
                self.assertTrue(
                    torch.all(periodic.period_length == fixed_period_length)
                )
        # test no temporal features
        msg = "The temporal kernel should only be used if there are temporal features."
        with self.assertRaisesRegex(AxError, msg):
            TemporalKernel(
                dim=10,
                matern_ard_num_dims=10,
                temporal_features=[],
                batch_shape=torch.Size([]),
                lengthscale_prior=ls_prior,
                outputscale_prior=os_prior,
                temporal_lengthscale_prior=tls_prior,
                period_length_prior=pl_prior,
                lengthscale_constraint=ls_constraint,
                outputscale_constraint=os_constraint,
                period_length_constraint=pl_constraint,
                temporal_lengthscale_constraint=tls_constraint,
            )
        # test exception raise when fixed_period_length is not None
        # and period_length_prior or period_length_constraint are provided
        msg = (
            "If `fixed_period_length` is provided, then `period_length_prior` "
            "and `period_length_constraint` are not used."
        )
        with self.assertRaisesRegex(ValueError, msg):
            TemporalKernel(
                dim=10,
                matern_ard_num_dims=8,
                temporal_features=[8, 9],
                batch_shape=torch.Size([]),
                lengthscale_prior=ls_prior,
                outputscale_prior=os_prior,
                fixed_period_length=5.0,
                temporal_lengthscale_prior=tls_prior,
                period_length_prior=pl_prior,
                lengthscale_constraint=ls_constraint,
                outputscale_constraint=os_constraint,
                temporal_lengthscale_constraint=tls_constraint,
            )
        with self.assertRaisesRegex(ValueError, msg):
            TemporalKernel(
                dim=10,
                matern_ard_num_dims=8,
                temporal_features=[8, 9],
                batch_shape=torch.Size([]),
                lengthscale_prior=ls_prior,
                outputscale_prior=os_prior,
                fixed_period_length=5.0,
                temporal_lengthscale_prior=tls_prior,
                period_length_constraint=pl_constraint,
                lengthscale_constraint=ls_constraint,
                outputscale_constraint=os_constraint,
                temporal_lengthscale_constraint=tls_constraint,
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.analysis.parallel_coordinates_plot import ParallelCoordinatesPlot
from ax.core.batch_trial import BatchTrial

from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_data_batch, get_branin_experiment
from plotly import graph_objs as go


class TestParallelCoordinatesPlot(TestCase):
    def test_parallel_coordinates_plot(self) -> None:
        exp = get_branin_experiment(with_completed_trial=True)

        parallel_coordinates_plot = ParallelCoordinatesPlot(
            experiment=exp, objective_name="branin"
        )

        # Assert that plot can be constructed successfully
        plot = parallel_coordinates_plot.get_fig()
        self.assertIsInstance(plot, go.Figure)

    def test_parallel_coordinates_plot_batch_trial(self) -> None:
        exp = get_branin_experiment(with_batch=True)

        batch_trial = exp.trials[0]
        if not isinstance(batch_trial, BatchTrial):
            raise TypeError("was not a batch trial")
        batch_trial.mark_running(no_runner_required=True)

        exp.attach_data(
            get_branin_data_batch(batch=batch_trial)
        )  # Add data for batch trial
        batch_trial.mark_completed()

        parallel_coordinates_plot = ParallelCoordinatesPlot(
            experiment=exp, objective_name="branin"
        )

        # Assert that plot can be constructed successfully
        plot = parallel_coordinates_plot.get_fig()
        self.assertIsInstance(plot, go.Figure)

    def test_parallel_coordinates_plot_objective_from_config(self) -> None:
        exp = get_branin_experiment(
            has_optimization_config=True, with_completed_trial=True
        )

        parallel_coordinates_plot = ParallelCoordinatesPlot(experiment=exp)

        # Assert that plot can be constructed successfully
        plot = parallel_coordinates_plot.get_fig()
        self.assertIsInstance(plot, go.Figure)

    def test_metric_not_found(self) -> None:
        exp = get_branin_experiment(with_completed_trial=True)

        parallel_coordinates_plot = ParallelCoordinatesPlot(
            experiment=exp, objective_name="invalid_metric"
        )

        # Assert that plot can be constructed successfully

        with self.assertRaises(ValueError):
            _ = parallel_coordinates_plot.get_df()

    def test_parallel_coordinates_get_df(self) -> None:
        exp = get_branin_experiment(with_completed_trial=True)

        parallel_coordinates_plot = ParallelCoordinatesPlot(
            experiment=exp, objective_name="branin"
        )

        # Assert that dataframe can be constructed successfully
        df = parallel_coordinates_plot.get_df()
        # check objective in dataframe
        self.assertIn("branin", df.columns)
        # check parameter names in dataframe
        self.assertIn("x1", df.columns)
        self.assertIn("x2", df.columns)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from typing import Any, Optional

import numpy as np
import pandas as pd
from ax.benchmark.metrics.base import BenchmarkMetricBase, GroundTruthMetricMixin
from ax.core.base_trial import BaseTrial
from ax.core.data import Data
from ax.core.metric import MetricFetchE, MetricFetchResult
from ax.utils.common.result import Err, Ok
from ax.utils.common.typeutils import not_none


class JenattonMetric(BenchmarkMetricBase):
    """Jenatton metric for hierarchical search spaces."""

    has_ground_truth: bool = True

    def __init__(
        self,
        name: str = "jenatton",
        noise_std: float = 0.0,
        observe_noise_sd: bool = False,
    ) -> None:
        super().__init__(name=name)
        self.noise_std = noise_std
        self.observe_noise_sd = observe_noise_sd
        self.lower_is_better = True

    def fetch_trial_data(self, trial: BaseTrial, **kwargs: Any) -> MetricFetchResult:
        try:
            mean = [
                jenatton_test_function(**arm.parameters)  # pyre-ignore [6]
                for _, arm in trial.arms_by_name.items()
            ]
            if self.noise_std != 0:
                mean = [m + self.noise_std * np.random.randn() for m in mean]
            df = pd.DataFrame(
                {
                    "arm_name": [name for name, _ in trial.arms_by_name.items()],
                    "metric_name": self.name,
                    "mean": mean,
                    "sem": self.noise_std if self.observe_noise_sd else None,
                    "trial_index": trial.index,
                }
            )
            return Ok(value=Data(df=df))

        except Exception as e:
            return Err(
                MetricFetchE(message=f"Failed to fetch {self.name}", exception=e)
            )

    def make_ground_truth_metric(self) -> GroundTruthJenattonMetric:
        return GroundTruthJenattonMetric(original_metric=self)


class GroundTruthJenattonMetric(JenattonMetric, GroundTruthMetricMixin):
    def __init__(self, original_metric: JenattonMetric) -> None:
        """
        Args:
            original_metric: The original JenattonMetric to which this metric
                corresponds.
        """
        super().__init__(
            name=self.get_ground_truth_name(original_metric),
            noise_std=0.0,
            observe_noise_sd=False,
        )


def jenatton_test_function(
    x1: Optional[int] = None,
    x2: Optional[int] = None,
    x3: Optional[int] = None,
    x4: Optional[float] = None,
    x5: Optional[float] = None,
    x6: Optional[float] = None,
    x7: Optional[float] = None,
    r8: Optional[float] = None,
    r9: Optional[float] = None,
) -> float:
    """Jenatton test function for hierarchical search spaces.

    This function is taken from:

    R. Jenatton, C. Archambeau, J. González, and M. Seeger. Bayesian
    optimization with tree-structured dependencies. ICML 2017.
    """
    if x1 == 0:
        if x2 == 0:
            return not_none(x4) ** 2 + 0.1 + not_none(r8)
        else:
            return not_none(x5) ** 2 + 0.2 + not_none(r8)
    else:
        if x3 == 0:
            return not_none(x6) ** 2 + 0.3 + not_none(r9)
        else:
            return not_none(x7) ** 2 + 0.4 + not_none(r9)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import itertools
import logging
from functools import reduce
from operator import mul
from typing import Dict, List, Optional, Tuple

import numpy as np
from ax.core.types import TGenMetadata, TParamValue, TParamValueList
from ax.models.discrete_base import DiscreteModel
from ax.models.types import TConfig
from ax.utils.common.docutils import copy_doc
from ax.utils.common.logger import get_logger


logger: logging.Logger = get_logger(__name__)


class FullFactorialGenerator(DiscreteModel):
    """Generator for full factorial designs.

    Generates arms for all possible combinations of parameter values,
    each with weight 1.

    The value of n supplied to `gen` will be ignored, as the number
    of arms generated is determined by the list of parameter values.
    To suppress this warning, use n = -1.
    """

    def __init__(
        self, max_cardinality: int = 100, check_cardinality: bool = True
    ) -> None:
        """
        Args:
            max_cardinality: maximum number of arms allowed if
                check_cardinality == True. Default is 100.
            check_cardinality: if True, throw if number of arms
                exceeds max_cardinality.
        """
        super().__init__()
        self.max_cardinality = max_cardinality
        self.check_cardinality = check_cardinality

    @copy_doc(DiscreteModel.gen)
    def gen(
        self,
        n: int,
        parameter_values: List[TParamValueList],
        objective_weights: Optional[np.ndarray],
        outcome_constraints: Optional[Tuple[np.ndarray, np.ndarray]] = None,
        fixed_features: Optional[Dict[int, TParamValue]] = None,
        pending_observations: Optional[List[List[TParamValueList]]] = None,
        model_gen_options: Optional[TConfig] = None,
    ) -> Tuple[List[TParamValueList], List[float], TGenMetadata]:
        if n != -1:
            raise ValueError(
                "FullFactorialGenerator will ignore the specified value of n. "
                "The generator automatically determines how many arms to "
                "generate."
            )

        if fixed_features:
            for fixed_feature_index, fixed_feature_value in fixed_features.items():
                parameter_values[fixed_feature_index] = [fixed_feature_value]

        num_arms = reduce(mul, [len(values) for values in parameter_values], 1)
        if self.check_cardinality and num_arms > self.max_cardinality:
            raise ValueError(
                f"FullFactorialGenerator generated {num_arms} arms, "
                f"but the maximum number of arms allowed is "
                f"{self.max_cardinality}."
            )

        points = [list(x) for x in itertools.product(*parameter_values)]
        return (points, [1.0 for _ in range(len(points))], {})



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections import defaultdict
from logging import Logger
from typing import Dict, List, Optional, Tuple

import pandas as pd
from ax.core.base_trial import TrialStatus
from ax.core.experiment import Experiment
from ax.core.map_data import MapData
from ax.utils.common.logger import get_logger
from pyre_extensions import assert_is_instance

logger: Logger = get_logger(__name__)


def align_partial_results(
    df: pd.DataFrame,
    progr_key: str,  # progression key
    metrics: List[str],
    interpolation: str = "slinear",
    do_forward_fill: bool = False,
    # TODO: Allow normalizing progr_key (e.g. subtract min time stamp)
) -> Tuple[Dict[str, pd.DataFrame], Dict[str, pd.DataFrame]]:
    """Helper function to align partial results with heterogeneous index

    Args:
        df: The DataFrame containing the raw data (in long format).
        progr_key: The key of the column indexing progression (such as
            the number of training examples, timestamps, etc.).
        metrics: The names of the metrics to consider.
        interpolation: The interpolation method used to fill missing values
            (if applicable). See `pandas.DataFrame.interpolate` for
            available options. Limit area is `inside`.
        forward_fill: If True, performs a forward fill after interpolation.
            This is useful for scalarizing learning curves when some data
            is missing. For instance, suppose we obtain a curve for task_1
            for progression in [a, b] and task_2 for progression in [c, d]
            where b < c. Performing the forward fill on task_1 is a possible
            solution.

    Returns:
        A two-tuple containing a dict mapping the provided metric names to the
        index-normalized and interpolated mean (sem).
    """
    missing_metrics = set(metrics) - set(df["metric_name"])
    if missing_metrics:
        raise ValueError(f"Metrics {missing_metrics} not found in input dataframe")
    # select relevant metrics
    df = df[df["metric_name"].isin(metrics)]
    # log some information about raw data
    for m in metrics:
        df_m = df[df["metric_name"] == m]
        if len(df_m) > 0:
            logger.debug(
                f"Metric {m} raw data has observations from "
                f"{df_m[progr_key].min()} to {df_m[progr_key].max()}."
            )
        else:
            logger.info(f"No data from metric {m} yet.")
    # drop arm names (assumes 1:1 map between trial indices and arm names)
    df = df.drop("arm_name", axis=1)
    # remove duplicates (same trial, metric, progr_key), which can happen
    # if the same progression is erroneously reported more than once
    df = df.drop_duplicates(
        subset=["trial_index", "metric_name", progr_key], keep="first"
    )
    # set multi-index over trial, metric, and progression key
    df = df.set_index(["trial_index", "metric_name", progr_key])
    # sort index
    df = df.sort_index()
    # drop sem if all NaN (assumes presence of sem column)
    has_sem = not df["sem"].isnull().all()
    if not has_sem:
        df = df.drop("sem", axis=1)
    # create the common index that every map result will be re-indexed w.r.t.
    index_union = df.index.levels[2].unique()
    # loop through (trial, metric) combos and align data
    dfs_mean = defaultdict(list)
    dfs_sem = defaultdict(list)
    for tidx in df.index.levels[0]:  # this could be slow if there are many trials
        for metric in df.index.levels[1]:
            # grab trial+metric sub-df and reindex to common index
            df_ridx = df.loc[(tidx, metric)].reindex(index_union)
            # interpolate / fill missing results
            # TODO: Allow passing of additional kwargs to `interpolate`
            # TODO: Allow using an arbitrary prediction model for this instead
            try:
                df_interp = df_ridx.interpolate(
                    method=interpolation, limit_area="inside"
                )
                if do_forward_fill:
                    # do forward fill (with valid observations) to handle instances
                    # where one task only has data for early progressions
                    df_interp = df_interp.fillna(method="pad")
            except ValueError as e:
                df_interp = df_ridx
                logger.info(
                    f"Got exception `{e}` during interpolation. "
                    "Using uninterpolated values instead."
                )
            # renaming column to trial index, append results
            dfs_mean[metric].append(df_interp["mean"].rename(tidx))
            if has_sem:
                dfs_sem[metric].append(df_interp["sem"].rename(tidx))

    # combine results into output dataframes
    dfs_mean = {metric: pd.concat(dfs, axis=1) for metric, dfs in dfs_mean.items()}
    dfs_sem = {metric: pd.concat(dfs, axis=1) for metric, dfs in dfs_sem.items()}

    return dfs_mean, dfs_sem


def estimate_early_stopping_savings(
    experiment: Experiment,
    map_key: Optional[str] = None,
) -> float:
    """Estimate resource savings due to early stopping by considering
    COMPLETED and EARLY_STOPPED trials. First, use the mean of final
    progressions of the set completed trials as a benchmark for the
    length of a single trial. The savings is then estimated as:

    resource_savings =
      1 - actual_resource_usage / (num_trials * length of single trial)

    Args:
        experiment: The experiment.
        map_key: The map_key to use when computing resource savings.

    Returns:
        The estimated resource savings as a fraction of total resource usage (i.e.
        0.11 estimated savings indicates we would expect the experiment to have used 11%
        more resources without early stopping present).
    """

    map_data = assert_is_instance(experiment.lookup_data(), MapData)

    # If no map_key is provided, use some arbitrary map_key in the experiment's MapData
    if map_key is not None:
        step_key = map_key
    elif len(map_data.map_key_infos) > 0:
        step_key = map_data.map_key_infos[0].key
    else:
        return 0

    # Get final number of steps of each trial
    trial_resources = (
        map_data.map_df[["trial_index", step_key]]
        .groupby("trial_index")
        .max()
        .reset_index()
    )

    early_stopped_trial_idcs = experiment.trial_indices_by_status[
        TrialStatus.EARLY_STOPPED
    ]
    completed_trial_idcs = experiment.trial_indices_by_status[TrialStatus.COMPLETED]

    # Assume that any early stopped trial would have had the mean number of steps of
    # the completed trials
    mean_completed_trial_resources = trial_resources[
        trial_resources["trial_index"].isin(completed_trial_idcs)
    ][step_key].mean()

    # Calculate the steps saved per early stopped trial. If savings are estimated to be
    # negative assume no savings
    stopped_trial_resources = trial_resources[
        trial_resources["trial_index"].isin(early_stopped_trial_idcs)
    ][step_key]
    saved_trial_resources = (
        mean_completed_trial_resources - stopped_trial_resources
    ).clip(0)

    # Return the ratio of the total saved resources over the total resources used plus
    # the total saved resources
    return saved_trial_resources.sum() / trial_resources[step_key].sum()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import warnings
from typing import Any, List, Optional

from ax.core.experiment import Experiment
from ax.core.search_space import SearchSpace
from ax.exceptions.storage import SQADecodeError
from ax.utils.common.base import Base, SortableBase


JSON_ATTRS = ["baseline_workflow_inputs"]
# Skip over the following attrs in `copy_db_ids`:
# * _experiment (to prevent infinite loops)
# * most generator run and generation strategy metadata
#   (since no Base objects are nested in there,
#   and we don't have guarantees about the structure of some
#   of that data, so the recursion could fail somewhere)
COPY_DB_IDS_ATTRS_TO_SKIP = {
    "_best_arm_predictions",
    "_bridge_kwargs",
    "_candidate_metadata_by_arm_signature",
    "_curr",
    "_experiment",
    "_gen_metadata",
    "_memo_df",
    "_model_kwargs",
    "_model_predictions",
    "_model_state_after_gen",
    "_model",
    "_seen_trial_indices_by_status",
    "_steps",
    "analysis_scheduler",
    "_nodes",
    # ``status_quo_weight_override`` is a field on ``BatchTrial`` not in the
    # "trial_v2" table
    # TODO(T193258337)
    "_status_quo_weight_override",
}
SKIP_ATTRS_ERROR_SUFFIX = "Consider adding to COPY_DB_IDS_ATTRS_TO_SKIP if appropriate."


def is_foreign_key_field(field: str) -> bool:
    """Return true if field name is a foreign key field, i.e. ends in `_id`."""
    return len(field) > 3 and field[-3:] == "_id"


# pyre-fixme[2]: Parameter annotation cannot be `Any`.
def copy_db_ids(source: Any, target: Any, path: Optional[List[str]] = None) -> None:
    """Takes as input two objects, `source` and `target`, that should be identical,
    except that `source` has _db_ids set and `target` doesn't. Recursively copies the
    _db_ids from `source` to `target`.

    Raise a SQADecodeError when the assumption of equality on `source` and `target`
    is violated, since this method is meant to be used when returning a new
    user-facing object after saving.
    """
    if not path:
        path = []

    error_message_prefix = (
        f"Error encountered while traversing source {path + [str(source)]} and "
        f"target {path + [str(target)]}: "
    )

    if len(path) > 15:
        # This shouldn't happen, but is a precaution against accidentally
        # introducing infinite loops
        raise SQADecodeError(error_message_prefix + "Encountered path of length > 10.")

    if type(source) is not type(target):
        if not issubclass(type(target), type(source)):
            if source is None and isinstance(target, SearchSpace):
                warnings.warn(
                    error_message_prefix + "Encountered two objects of different "
                    f"types: {type(source)} and {type(target)}. Continuing in the "
                    "special case that `source is None` and `target` is a "
                    "`SearchSpace`."
                )
            else:
                raise SQADecodeError(
                    error_message_prefix + "Encountered two objects of different "
                    f"types: {type(source)} and {type(target)}."
                    + SKIP_ATTRS_ERROR_SUFFIX
                )

    if isinstance(source, Base):
        for attr, val in source.__dict__.items():
            if attr.endswith("_db_id"):
                # we're at a "leaf" node; copy the db_id and return
                setattr(target, attr, val)
                continue

            # Skip attrs that are doubly private or in COPY_DB_IDS_TO_SKIP.
            if attr.startswith("__") or attr in COPY_DB_IDS_ATTRS_TO_SKIP:
                continue

            # For Json attributes we would like to simply test for equality and not
            # recurse through the Json.
            # TODO: Add json_attrs as an argument and plumb through to `copy_db_ids`.
            if attr in JSON_ATTRS:
                source_json = getattr(source, attr)
                target_json = getattr(target, attr)
                if source_json != target_json:
                    SQADecodeError(
                        error_message_prefix + f"Json attribute {attr} not matching "
                        f"between source: {source_json} and target: {target_json}."
                    )
                continue

            # Arms are referenced twice on an Experiment object; once in
            # experiment.arms_by_name/signature and once in
            # trial.arms_by_name/signature. When copying db_ids, we should
            # ignore the former, since it will "collapse" arms of the same
            # name/signature that appear in more than one trial.
            if isinstance(source, Experiment) and attr in {
                "_arms_by_name",
                "_arms_by_signature",
            }:
                continue

            copy_db_ids(val, getattr(target, attr), path + [attr])

    elif isinstance(source, (list, set)):
        source = list(source)
        target = list(target)

        if len(source) != len(target):
            raise SQADecodeError(
                error_message_prefix + "Encountered lists of different lengths."
            )

        if len(source) == 0:
            return

        if isinstance(source[0], Base) and not isinstance(source[0], SortableBase):
            raise SQADecodeError(
                error_message_prefix + f"Cannot sort instances of {type(source[0])}; "
                "sorting is only defined on instances of SortableBase."
            )

        try:
            source = sorted(source)
            target = sorted(target)
        except TypeError as e:
            if any(isinstance(o, Base) for o in source + target):
                raise SQADecodeError(
                    error_message_prefix + f"TypeError encountered during sorting: {e}"
                )
            else:
                # source and target are not lists of things that need to be saved
                return

        for index, x in enumerate(source):
            copy_db_ids(x, target[index], path + [str(index)])

    elif isinstance(source, dict):
        for k, v in source.items():
            if k not in target:
                raise SQADecodeError(
                    error_message_prefix + "Encountered key only present "
                    f"in source dictionary: {k}."
                )
            copy_db_ids(v, target[k], path + [k])

    else:
        return



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from typing import cast

import torch
from ax.core.search_space import SearchSpaceDigest
from ax.models.torch.cbo_lcea import LCEABO
from ax.utils.common.testutils import TestCase
from ax.utils.testing.mock import fast_botorch_optimize
from botorch.models.contextual import LCEAGP
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.utils.datasets import SupervisedDataset


class LCEABOTest(TestCase):
    @fast_botorch_optimize
    def test_LCEABO(self) -> None:
        train_X = torch.tensor(
            [[0.0, 0.0, 0.0, 0.0], [1.0, 1.0, 1.0, 1.0], [2.0, 2.0, 2.0, 2.0]]
        )
        train_Y = torch.tensor([[1.0], [2.0], [3.0]])
        train_Yvar = 0.1 * torch.ones(3, 1)
        feature_names = ["x0", "x1", "x2", "x3"]
        metric_names = ["y"]
        training_data = [
            SupervisedDataset(
                X=train_X,
                Y=train_Y,
                Yvar=train_Yvar,
                feature_names=feature_names,
                outcome_names=metric_names,
            )
        ]

        # Test setting attributes
        decomposition = {"1": ["x0", "x1"], "2": ["x2", "x3"]}
        m1 = LCEABO(decomposition=decomposition)
        self.assertDictEqual(m1.decomposition, decomposition)
        self.assertTrue(m1.train_embedding)
        self.assertIsNone(m1.cat_feature_dict)
        self.assertIsNone(m1.embs_feature_dict)
        self.assertIsNone(m1.embs_dim_list)

        # Test fit
        m1.fit(
            datasets=training_data,
            search_space_digest=SearchSpaceDigest(
                feature_names=feature_names,
                bounds=[(0.0, 1.0) for _ in range(4)],
            ),
        )
        self.assertIsInstance(m1.model, LCEAGP)

        # Test get_and_fit_model with single metric
        gp = m1.get_and_fit_model(
            Xs=[train_X],
            Ys=[train_Y],
            Yvars=[train_Yvar],
            task_features=[],
            fidelity_features=[],
            metric_names=["y"],
        )
        self.assertIsInstance(gp, LCEAGP)

        # Test get_and_fit_model with multiple metrics
        gp_list = m1.get_and_fit_model(
            Xs=[train_X, train_X],
            Ys=[train_Y, train_Y],
            Yvars=[train_Yvar, train_Yvar],
            task_features=[],
            fidelity_features=[],
            metric_names=metric_names,
        )
        self.assertIsInstance(gp_list, ModelListGP)

        # Test decomposition validation in __init__
        with self.assertRaises(AssertionError):
            LCEABO(decomposition={"1": ["x0"], "2": ["x1", "x3"]})

        # Test input decomposition indicates parameter name
        m2 = LCEABO(decomposition={"1": ["x0", "x2"], "2": ["x1", "x3"]})
        m2.fit(
            datasets=training_data,
            search_space_digest=SearchSpaceDigest(
                feature_names=feature_names,
                bounds=[(0.0, 1.0) for _ in range(4)],
            ),
        )
        self.assertDictEqual(
            cast(LCEAGP, m2.model).decomposition, {"1": [0, 2], "2": [1, 3]}
        )

        # Test decomposition validation in get_and_fit_model
        # does not pass feature names when decomposition uses feature names
        with self.assertRaises(ValueError):
            m2.fit(
                datasets=training_data,
                search_space_digest=SearchSpaceDigest(
                    feature_names=[],
                    bounds=[(0.0, 1.0) for _ in range(4)],
                ),
            )

        # pass wrong feature names
        with self.assertRaises(AssertionError):
            m2.fit(
                datasets=training_data,
                search_space_digest=SearchSpaceDigest(
                    feature_names=["x0", "x6", "x10", "z"],
                    bounds=[(0.0, 1.0) for _ in range(4)],
                ),
            )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger

from ax.plot.base import AxPlotConfig, AxPlotTypes
from ax.plot.render import _js_requires, _wrap_js, plot_config_to_html
from ax.utils.common.logger import get_logger
from IPython.display import display
from plotly.offline import init_notebook_mode, iplot

logger: Logger = get_logger(__name__)


def init_notebook_plotting(offline: bool = False) -> None:
    """Initialize plotting in notebooks, either in online or offline mode."""
    display_bundle = {"text/html": _wrap_js(_js_requires(offline=offline))}
    display(display_bundle, raw=True)
    logger.info("Injecting Plotly library into cell. Do not overwrite or delete cell.")
    logger.info(
        """
    Please see
    (https://ax.dev/tutorials/visualizations.html#Fix-for-plots-that-are-not-rendering)
    if visualizations are not rendering.
    """.strip()
    )
    init_notebook_mode()


def render(plot_config: AxPlotConfig, inject_helpers: bool = False) -> None:
    """Render plot config."""
    if plot_config.plot_type == AxPlotTypes.GENERIC:
        iplot(plot_config.data)
    elif plot_config.plot_type == AxPlotTypes.HTML:
        assert "text/html" in plot_config.data
        display(plot_config.data, raw=True)
    else:
        display_bundle = {
            "text/html": plot_config_to_html(plot_config, inject_helpers=inject_helpers)
        }
        display(display_bundle, raw=True)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any, List

import pandas as pd
import plotly.graph_objs as go
from ax.modelbridge.base import ModelBridge
from ax.plot.base import AxPlotConfig, AxPlotTypes, DECIMALS
from ax.plot.helper import get_plot_data
from ax.utils.stats.statstools import marginal_effects
from plotly import subplots


def plot_marginal_effects(model: ModelBridge, metric: str) -> AxPlotConfig:
    """
    Calculates and plots the marginal effects -- the effect of changing one
    factor away from the randomized distribution of the experiment and fixing it
    at a particular level.

    Args:
        model: Model to use for estimating effects
        metric: The metric for which to plot marginal effects.

    Returns:
        AxPlotConfig of the marginal effects
    """
    plot_data, _, _ = get_plot_data(model, {}, {metric})

    arm_dfs = []
    for arm in plot_data.in_sample.values():
        arm_df = pd.DataFrame(arm.parameters, index=[arm.name])
        arm_df["mean"] = arm.y_hat[metric]
        arm_df["sem"] = arm.se_hat[metric]
        arm_dfs.append(arm_df)
    effect_table = marginal_effects(pd.concat(arm_dfs, axis=0))

    varnames = effect_table["Name"].unique()
    # pyre-fixme[33]: Given annotation cannot contain `Any`.
    data: List[Any] = []
    for varname in varnames:
        var_df = effect_table[effect_table["Name"] == varname]
        data += [
            go.Bar(
                x=var_df["Level"],
                y=var_df["Beta"],
                error_y={"type": "data", "array": var_df["SE"]},
                name=varname,
            )
        ]
    fig = subplots.make_subplots(
        cols=len(varnames),
        rows=1,
        subplot_titles=list(varnames),
        print_grid=False,
        shared_yaxes=True,
    )
    for idx, item in enumerate(data):
        fig.append_trace(item, 1, idx + 1)  # pyre-ignore[16]
    fig.layout.showlegend = False
    # fig.layout.margin = go.layout.Margin(l=2, r=2)
    fig.layout.title = "Marginal Effects by Factor"
    fig.layout.yaxis = {
        "title": "% higher than experiment average",
        "hoverformat": ".{}f".format(DECIMALS),
    }
    return AxPlotConfig(data=fig, plot_type=AxPlotTypes.GENERIC)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger
from typing import Dict, List, Optional, Type

import torch
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.optimization_config import OptimizationConfig
from ax.core.search_space import SearchSpace
from ax.modelbridge.discrete import DiscreteModelBridge
from ax.modelbridge.random import RandomModelBridge
from ax.modelbridge.registry import Cont_X_trans, Models, Y_trans
from ax.modelbridge.torch import TorchModelBridge
from ax.modelbridge.transforms.base import Transform
from ax.models.torch.botorch import (
    TAcqfConstructor,
    TModelConstructor,
    TModelPredictor,
    TOptimizer,
)
from ax.models.torch.botorch_defaults import (
    get_and_fit_model,
    get_qLogNEI,
    scipy_optimizer,
)
from ax.models.torch.utils import predict_from_model
from ax.models.types import TConfig
from ax.utils.common.logger import get_logger
from ax.utils.common.typeutils import checked_cast


logger: Logger = get_logger(__name__)


DEFAULT_TORCH_DEVICE = torch.device("cpu")
DEFAULT_EHVI_BATCH_LIMIT = 5


"""
Module containing functions that generate standard models, such as Sobol,
GP+EI, etc.

Note: a special case here is a composite generator, which requires an
additional ``GenerationStrategy`` and is able to delegate work to multiple models
(for instance, to a random model to generate the first trial, and to an
optimization model for subsequent trials).

"""


def get_sobol(
    search_space: SearchSpace,
    seed: Optional[int] = None,
    deduplicate: bool = False,
    init_position: int = 0,
    scramble: bool = True,
    fallback_to_sample_polytope: bool = False,
) -> RandomModelBridge:
    """Instantiates a Sobol sequence quasi-random generator.

    Args:
        search_space: Sobol generator search space.
        kwargs: Custom args for sobol generator.

    Returns:
        RandomModelBridge, with SobolGenerator as model.
    """
    return checked_cast(
        RandomModelBridge,
        Models.SOBOL(
            search_space=search_space,
            seed=seed,
            deduplicate=deduplicate,
            init_position=init_position,
            scramble=scramble,
            fallback_to_sample_polytope=fallback_to_sample_polytope,
        ),
    )


def get_uniform(
    search_space: SearchSpace, deduplicate: bool = False, seed: Optional[int] = None
) -> RandomModelBridge:
    """Instantiate uniform generator.

    Args:
        search_space: Uniform generator search space.
        kwargs: Custom args for uniform generator.

    Returns:
        RandomModelBridge, with UniformGenerator as model.
    """
    return checked_cast(
        RandomModelBridge,
        Models.UNIFORM(search_space=search_space, seed=seed, deduplicate=deduplicate),
    )


def get_botorch(
    experiment: Experiment,
    data: Data,
    search_space: Optional[SearchSpace] = None,
    dtype: torch.dtype = torch.double,
    device: torch.device = DEFAULT_TORCH_DEVICE,
    transforms: List[Type[Transform]] = Cont_X_trans + Y_trans,
    transform_configs: Optional[Dict[str, TConfig]] = None,
    model_constructor: TModelConstructor = get_and_fit_model,
    model_predictor: TModelPredictor = predict_from_model,
    acqf_constructor: TAcqfConstructor = get_qLogNEI,
    acqf_optimizer: TOptimizer = scipy_optimizer,  # pyre-ignore[9]
    refit_on_cv: bool = False,
    optimization_config: Optional[OptimizationConfig] = None,
) -> TorchModelBridge:
    """Instantiates a BotorchModel."""
    if data.df.empty:
        raise ValueError("`BotorchModel` requires non-empty data.")
    return checked_cast(
        TorchModelBridge,
        Models.LEGACY_BOTORCH(
            experiment=experiment,
            data=data,
            search_space=search_space or experiment.search_space,
            torch_dtype=dtype,
            torch_device=device,
            transforms=transforms,
            transform_configs=transform_configs,
            model_constructor=model_constructor,
            model_predictor=model_predictor,
            acqf_constructor=acqf_constructor,
            acqf_optimizer=acqf_optimizer,
            refit_on_cv=refit_on_cv,
            optimization_config=optimization_config,
        ),
    )


def get_GPEI(
    experiment: Experiment,
    data: Data,
    search_space: Optional[SearchSpace] = None,
    dtype: torch.dtype = torch.double,
    device: torch.device = DEFAULT_TORCH_DEVICE,
) -> TorchModelBridge:
    """Instantiates a GP model that generates points with EI."""
    if data.df.empty:
        raise ValueError("GP+EI BotorchModel requires non-empty data.")
    return checked_cast(
        TorchModelBridge,
        Models.LEGACY_BOTORCH(
            experiment=experiment,
            data=data,
            search_space=search_space or experiment.search_space,
            torch_dtype=dtype,
            torch_device=device,
        ),
    )


def get_factorial(search_space: SearchSpace) -> DiscreteModelBridge:
    """Instantiates a factorial generator."""
    return checked_cast(
        DiscreteModelBridge,
        Models.FACTORIAL(search_space=search_space, fit_out_of_design=True),
    )


def get_empirical_bayes_thompson(
    experiment: Experiment,
    data: Data,
    search_space: Optional[SearchSpace] = None,
    num_samples: int = 10000,
    min_weight: Optional[float] = None,
    uniform_weights: bool = False,
) -> DiscreteModelBridge:
    """Instantiates an empirical Bayes / Thompson sampling model."""
    if data.df.empty:
        raise ValueError("Empirical Bayes Thompson sampler requires non-empty data.")
    return checked_cast(
        DiscreteModelBridge,
        Models.EMPIRICAL_BAYES_THOMPSON(
            experiment=experiment,
            data=data,
            search_space=search_space or experiment.search_space,
            num_samples=num_samples,
            min_weight=min_weight,
            uniform_weights=uniform_weights,
            fit_out_of_design=True,
        ),
    )


def get_thompson(
    experiment: Experiment,
    data: Data,
    search_space: Optional[SearchSpace] = None,
    num_samples: int = 10000,
    min_weight: Optional[float] = None,
    uniform_weights: bool = False,
) -> DiscreteModelBridge:
    """Instantiates a Thompson sampling model."""
    if data.df.empty:
        raise ValueError("Thompson sampler requires non-empty data.")
    return checked_cast(
        DiscreteModelBridge,
        Models.THOMPSON(
            experiment=experiment,
            data=data,
            search_space=search_space or experiment.search_space,
            num_samples=num_samples,
            min_weight=min_weight,
            uniform_weights=uniform_weights,
            fit_out_of_design=True,
        ),
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Module containing a registry of standard models (and generators, samplers etc.)
such as Sobol generator, GP+EI, Thompson sampler, etc.

Use of `Models` enum allows for serialization and reinstantiation of models and
generation strategies from generator runs they produced. To reinstantiate a model
from generator run, use `get_model_from_generator_run` utility from this module.
"""


from __future__ import annotations

import warnings
from enum import Enum
from inspect import isfunction, signature
from logging import Logger
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Type

import torch
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.search_space import SearchSpace
from ax.modelbridge.base import ModelBridge
from ax.modelbridge.discrete import DiscreteModelBridge
from ax.modelbridge.random import RandomModelBridge
from ax.modelbridge.torch import TorchModelBridge
from ax.modelbridge.transforms.base import Transform
from ax.modelbridge.transforms.choice_encode import (
    ChoiceToNumericChoice,
    OrderedChoiceToIntegerRange,
)
from ax.modelbridge.transforms.convert_metric_names import ConvertMetricNames
from ax.modelbridge.transforms.derelativize import Derelativize
from ax.modelbridge.transforms.int_range_to_choice import IntRangeToChoice
from ax.modelbridge.transforms.int_to_float import IntToFloat
from ax.modelbridge.transforms.ivw import IVW
from ax.modelbridge.transforms.log import Log
from ax.modelbridge.transforms.logit import Logit
from ax.modelbridge.transforms.one_hot import OneHot
from ax.modelbridge.transforms.relativize import Relativize
from ax.modelbridge.transforms.remove_fixed import RemoveFixed
from ax.modelbridge.transforms.search_space_to_choice import SearchSpaceToChoice
from ax.modelbridge.transforms.standardize_y import StandardizeY
from ax.modelbridge.transforms.stratified_standardize_y import StratifiedStandardizeY
from ax.modelbridge.transforms.task_encode import TaskChoiceToIntTaskChoice
from ax.modelbridge.transforms.trial_as_task import TrialAsTask
from ax.modelbridge.transforms.unit_x import UnitX
from ax.models.base import Model
from ax.models.discrete.eb_thompson import EmpiricalBayesThompsonSampler
from ax.models.discrete.full_factorial import FullFactorialGenerator
from ax.models.discrete.thompson import ThompsonSampler
from ax.models.random.sobol import SobolGenerator
from ax.models.random.uniform import UniformGenerator
from ax.models.torch.botorch import BotorchModel
from ax.models.torch.botorch_modular.model import (
    BoTorchModel as ModularBoTorchModel,
    SurrogateSpec,
)
from ax.models.torch.botorch_moo import MultiObjectiveBotorchModel
from ax.models.torch.cbo_sac import SACBO
from ax.utils.common.kwargs import (
    consolidate_kwargs,
    get_function_argument_names,
    get_function_default_arguments,
    validate_kwarg_typing,
)
from ax.utils.common.logger import get_logger
from ax.utils.common.serialization import callable_from_reference, callable_to_reference
from ax.utils.common.typeutils import checked_cast, not_none
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.fully_bayesian_multitask import SaasFullyBayesianMultiTaskGP

logger: Logger = get_logger(__name__)

Cont_X_trans: List[Type[Transform]] = [
    RemoveFixed,
    OrderedChoiceToIntegerRange,
    OneHot,
    IntToFloat,
    Log,
    Logit,
    UnitX,
]

Discrete_X_trans: List[Type[Transform]] = [IntRangeToChoice]

Mixed_transforms: List[Type[Transform]] = [
    RemoveFixed,
    ChoiceToNumericChoice,
    IntToFloat,
    Log,
    Logit,
    UnitX,
]

EB_ashr_trans: List[Type[Transform]] = [Relativize, IVW, SearchSpaceToChoice]

Y_trans: List[Type[Transform]] = [IVW, Derelativize, StandardizeY]

# Expected `List[Type[Transform]]` for 2nd anonymous parameter to
# call `list.__add__` but got `List[Type[SearchSpaceToChoice]]`.
TS_trans: List[Type[Transform]] = Y_trans + [SearchSpaceToChoice]

# Multi-type MTGP transforms
MT_MTGP_trans: List[Type[Transform]] = Cont_X_trans + [
    Derelativize,
    ConvertMetricNames,
    TrialAsTask,
    StratifiedStandardizeY,
    TaskChoiceToIntTaskChoice,
]

# Single-type MTGP transforms
ST_MTGP_trans: List[Type[Transform]] = Cont_X_trans + [
    Derelativize,
    TrialAsTask,
    StratifiedStandardizeY,
    TaskChoiceToIntTaskChoice,
]

# Single-type MTGP transforms
Specified_Task_ST_MTGP_trans: List[Type[Transform]] = Cont_X_trans + [
    Derelativize,
    StratifiedStandardizeY,
    TaskChoiceToIntTaskChoice,
]

STANDARD_TORCH_BRIDGE_KWARGS: Dict[str, Any] = {"torch_dtype": torch.double}


class ModelSetup(NamedTuple):
    """A model setup defines a coupled combination of a model, a model bridge,
    standard set of transforms, and standard model bridge keyword arguments.
    This coupled combination yields a given standard modeling strategy in Ax,
    such as BoTorch GP+EI, a Thompson sampler, or a Sobol quasirandom generator.
    """

    bridge_class: Type[ModelBridge]
    model_class: Type[Model]
    transforms: List[Type[Transform]]
    default_model_kwargs: Optional[Dict[str, Any]] = None
    standard_bridge_kwargs: Optional[Dict[str, Any]] = None
    not_saved_model_kwargs: Optional[List[str]] = None


"""A mapping of string keys that indicate a model, to the corresponding
model setup, which defines which model, model bridge, transforms, and
standard arguments a given model requires.
"""
MODEL_KEY_TO_MODEL_SETUP: Dict[str, ModelSetup] = {
    "BoTorch": ModelSetup(
        bridge_class=TorchModelBridge,
        model_class=ModularBoTorchModel,
        transforms=Cont_X_trans + Y_trans,
        standard_bridge_kwargs=STANDARD_TORCH_BRIDGE_KWARGS,
    ),
    "GPEI": ModelSetup(
        bridge_class=TorchModelBridge,
        model_class=BotorchModel,
        transforms=Cont_X_trans + Y_trans,
        standard_bridge_kwargs=STANDARD_TORCH_BRIDGE_KWARGS,
    ),
    "EB": ModelSetup(
        bridge_class=DiscreteModelBridge,
        model_class=EmpiricalBayesThompsonSampler,
        transforms=TS_trans,
    ),
    "Factorial": ModelSetup(
        bridge_class=DiscreteModelBridge,
        model_class=FullFactorialGenerator,
        transforms=Discrete_X_trans,
    ),
    "Thompson": ModelSetup(
        bridge_class=DiscreteModelBridge,
        model_class=ThompsonSampler,
        transforms=TS_trans,
    ),
    "Sobol": ModelSetup(
        bridge_class=RandomModelBridge,
        model_class=SobolGenerator,
        transforms=Cont_X_trans,
    ),
    "Uniform": ModelSetup(
        bridge_class=RandomModelBridge,
        model_class=UniformGenerator,
        transforms=Cont_X_trans,
    ),
    "MOO": ModelSetup(
        bridge_class=TorchModelBridge,
        model_class=MultiObjectiveBotorchModel,
        transforms=Cont_X_trans + Y_trans,
        standard_bridge_kwargs=STANDARD_TORCH_BRIDGE_KWARGS,
    ),
    "ST_MTGP_LEGACY": ModelSetup(
        bridge_class=TorchModelBridge,
        model_class=BotorchModel,
        transforms=ST_MTGP_trans,
        standard_bridge_kwargs=STANDARD_TORCH_BRIDGE_KWARGS,
    ),
    "ST_MTGP": ModelSetup(
        bridge_class=TorchModelBridge,
        model_class=ModularBoTorchModel,
        transforms=ST_MTGP_trans,
        standard_bridge_kwargs=STANDARD_TORCH_BRIDGE_KWARGS,
    ),
    "BO_MIXED": ModelSetup(
        bridge_class=TorchModelBridge,
        model_class=ModularBoTorchModel,
        transforms=Mixed_transforms + Y_trans,
        standard_bridge_kwargs=STANDARD_TORCH_BRIDGE_KWARGS,
    ),
    "SAASBO": ModelSetup(
        bridge_class=TorchModelBridge,
        model_class=ModularBoTorchModel,
        transforms=Cont_X_trans + Y_trans,
        default_model_kwargs={
            "surrogate_specs": {
                "SAASBO_Surrogate": SurrogateSpec(
                    botorch_model_class=SaasFullyBayesianSingleTaskGP
                )
            },
        },
        standard_bridge_kwargs=STANDARD_TORCH_BRIDGE_KWARGS,
    ),
    "SAAS_MTGP": ModelSetup(
        bridge_class=TorchModelBridge,
        model_class=ModularBoTorchModel,
        transforms=ST_MTGP_trans,
        default_model_kwargs={
            "surrogate_specs": {
                "SAAS_MTGP_Surrogate": SurrogateSpec(
                    botorch_model_class=SaasFullyBayesianMultiTaskGP
                )
            },
        },
        standard_bridge_kwargs=STANDARD_TORCH_BRIDGE_KWARGS,
    ),
    "ST_MTGP_NEHVI": ModelSetup(
        bridge_class=TorchModelBridge,
        model_class=MultiObjectiveBotorchModel,
        transforms=ST_MTGP_trans,
        standard_bridge_kwargs=STANDARD_TORCH_BRIDGE_KWARGS,
    ),
    "Contextual_SACBO": ModelSetup(
        bridge_class=TorchModelBridge,
        model_class=SACBO,
        transforms=Cont_X_trans + Y_trans,
        standard_bridge_kwargs=STANDARD_TORCH_BRIDGE_KWARGS,
    ),
}


class ModelRegistryBase(Enum):
    """Base enum that provides instrumentation of `__call__` on enum values,
    for enums that link their values to `ModelSetup`-s like `Models`.
    """

    @property
    def model_class(self) -> Type[Model]:
        """Type of `Model` used for the given model+bridge setup."""
        return MODEL_KEY_TO_MODEL_SETUP[self.value].model_class

    @property
    def model_bridge_class(self) -> Type[ModelBridge]:
        """Type of `ModelBridge` used for the given model+bridge setup."""
        return MODEL_KEY_TO_MODEL_SETUP[self.value].bridge_class

    def __call__(
        self,
        search_space: Optional[SearchSpace] = None,
        experiment: Optional[Experiment] = None,
        data: Optional[Data] = None,
        silently_filter_kwargs: bool = False,
        **kwargs: Any,
    ) -> ModelBridge:
        assert self.value in MODEL_KEY_TO_MODEL_SETUP, f"Unknown model {self.value}"
        # All model bridges require either a search space or an experiment.
        assert search_space or experiment, "Search space or experiment required."
        search_space = search_space or not_none(experiment).search_space
        model_setup_info = MODEL_KEY_TO_MODEL_SETUP[self.value]
        model_class = model_setup_info.model_class
        bridge_class = model_setup_info.bridge_class
        if not silently_filter_kwargs:
            validate_kwarg_typing(
                typed_callables=[model_class, bridge_class],
                search_space=search_space,
                experiment=experiment,
                data=data,
                **kwargs,
            )

        # Create model with consolidated arguments: defaults + passed in kwargs.
        model_kwargs = consolidate_kwargs(
            kwargs_iterable=[
                get_function_default_arguments(model_class),
                model_setup_info.default_model_kwargs,
                kwargs,
            ],
            keywords=get_function_argument_names(model_class),
        )
        model = model_class(**model_kwargs)

        # Create `ModelBridge`: defaults + standard kwargs + passed in kwargs.
        bridge_kwargs = consolidate_kwargs(
            kwargs_iterable=[
                get_function_default_arguments(bridge_class),
                model_setup_info.standard_bridge_kwargs,
                {"transforms": model_setup_info.transforms},
                kwargs,
            ],
            keywords=get_function_argument_names(
                function=bridge_class, omit=["experiment", "search_space", "data"]
            ),
        )

        # Create model bridge with the consolidated kwargs.
        model_bridge = bridge_class(
            search_space=search_space or not_none(experiment).search_space,
            experiment=experiment,
            data=data,
            model=model,
            **bridge_kwargs,
        )

        if model_setup_info.not_saved_model_kwargs:
            for key in model_setup_info.not_saved_model_kwargs:
                model_kwargs.pop(key, None)

        # Store all kwargs on model bridge, to be saved on generator run.
        model_bridge._set_kwargs_to_save(
            model_key=self.value,
            model_kwargs=_encode_callables_as_references(model_kwargs),
            bridge_kwargs=_encode_callables_as_references(bridge_kwargs),
        )
        return model_bridge

    def view_defaults(self) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Obtains the default keyword arguments for the model and the modelbridge
        specified through the Models enum, for ease of use in notebook environment,
        since models and bridges cannot be inspected directly through the enum.

        Returns:
            A tuple of default keyword arguments for the model and the model bridge.
        """
        model_setup_info = not_none(MODEL_KEY_TO_MODEL_SETUP.get(self.value))
        return (
            self._get_model_kwargs(info=model_setup_info),
            self._get_bridge_kwargs(info=model_setup_info),
        )

    def view_kwargs(self) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Obtains annotated keyword arguments that the model and the modelbridge
        (corresponding to a given member of the Models enum) constructors expect.

        Returns:
            A tuple of annotated keyword arguments for the model and the model bridge.
        """
        model_class = self.model_class
        bridge_class = self.model_bridge_class
        return (
            {kw: p.annotation for kw, p in signature(model_class).parameters.items()},
            {kw: p.annotation for kw, p in signature(bridge_class).parameters.items()},
        )

    @staticmethod
    def _get_model_kwargs(
        info: ModelSetup, kwargs: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        return consolidate_kwargs(
            [get_function_default_arguments(info.model_class), kwargs],
            keywords=get_function_argument_names(info.model_class),
        )

    @staticmethod
    def _get_bridge_kwargs(
        info: ModelSetup, kwargs: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        return consolidate_kwargs(
            [
                get_function_default_arguments(info.bridge_class),
                info.standard_bridge_kwargs,
                {"transforms": info.transforms},
                kwargs,
            ],
            keywords=get_function_argument_names(
                info.bridge_class, omit=["experiment", "search_space", "data"]
            ),
        )


class Models(ModelRegistryBase):
    """Registry of available models.

    Uses MODEL_KEY_TO_MODEL_SETUP to retrieve settings for model and model bridge,
    by the key stored in the enum value.

    To instantiate a model in this enum, simply call an enum member like so:
    `Models.SOBOL(search_space=search_space)` or
    `Models.GPEI(experiment=experiment, data=data)`. Keyword arguments
    specified to the call will be passed into the model or the model bridge
    constructors according to their keyword.

    For instance, `Models.SOBOL(search_space=search_space, scramble=False)`
    will instantiate a `RandomModelBridge(search_space=search_space)`
    with a `SobolGenerator(scramble=False)` underlying model.
    """

    SOBOL = "Sobol"
    GPEI = "GPEI"
    FACTORIAL = "Factorial"
    SAASBO = "SAASBO"
    SAAS_MTGP = "SAAS_MTGP"
    THOMPSON = "Thompson"
    LEGACY_BOTORCH = "GPEI"
    BOTORCH_MODULAR = "BoTorch"
    EMPIRICAL_BAYES_THOMPSON = "EB"
    UNIFORM = "Uniform"
    MOO = "MOO"
    ST_MTGP_LEGACY = "ST_MTGP_LEGACY"
    ST_MTGP = "ST_MTGP"
    BO_MIXED = "BO_MIXED"
    ST_MTGP_NEHVI = "ST_MTGP_NEHVI"
    CONTEXT_SACBO = "Contextual_SACBO"

    @classmethod
    @property
    def BOTORCH(cls) -> Models:
        warnings.warn(
            "`BOTORCH` Model class has been deprecated and renamed to "
            "`LEGACY_BOTORCH`, and it will be removed in a future release. "
            "Please use `BOTORCH_MODULAR` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.LEGACY_BOTORCH

    @classmethod
    @property
    def FULLYBAYESIAN(cls) -> Models:
        return _deprecated_model_with_warning(
            old_model_str="FULLYBAYESIAN", new_model=cls.SAASBO
        )

    @classmethod
    @property
    def FULLYBAYESIANMOO(cls) -> Models:
        return _deprecated_model_with_warning(
            old_model_str="FULLYBAYESIANMOO", new_model=cls.SAASBO
        )

    @classmethod
    @property
    def FULLYBAYESIAN_MTGP(cls) -> Models:
        return _deprecated_model_with_warning(
            old_model_str="FULLYBAYESIAN_MTGP", new_model=cls.SAAS_MTGP
        )

    @classmethod
    @property
    def FULLYBAYESIANMOO_MTGP(cls) -> Models:
        return _deprecated_model_with_warning(
            old_model_str="FULLYBAYESIANMOO_MTGP", new_model=cls.SAAS_MTGP
        )


def _deprecated_model_with_warning(old_model_str: str, new_model: Models) -> Models:
    warnings.warn(
        f"{old_model_str} is deprecated and replaced by {new_model}. "
        f"Please use {new_model}. This will become an error in a future release.",
        DeprecationWarning,
        stacklevel=3,
    )
    return new_model


def get_model_from_generator_run(
    generator_run: GeneratorRun,
    experiment: Experiment,
    data: Data,
    models_enum: Type[ModelRegistryBase],
    after_gen: bool = True,
) -> ModelBridge:
    """Reinstantiate a model from model key and kwargs stored on a given generator
    run, with the given experiment and the data to initialize the model with.

    Note: requires that the model that was used to get the generator run, is part
    of the `Models` registry enum.

    Args:
        generator_run: A `GeneratorRun` created by the model we are looking to
            reinstantiate.
        experiment: The experiment for which the model is reinstantiated.
        data: Data, with which to reinstantiate the model.
        models_enum: Subclass of `Models` registry, from which to obtain
            the settings of the model. Useful only if the generator run was
            created via a model that could not be included into the main registry,
            but can still be represented as a `ModelSetup` and was added to a
            registry that extends `Models`.
        after_gen: Whether to reinstantiate the model in the state, in which it
            was after it created this generator run, as opposed to before.
            Defaults to True, useful when reinstantiating the model to resume
            optimization, rather than to recreate its state at the time of
            generation. TO recreate state at the time of generation, set to `False`.
    """
    if not generator_run._model_key:
        raise ValueError(
            "Cannot restore model from generator run as no model key was "
            "on the generator run stored."
        )
    model = models_enum(generator_run._model_key)
    model_kwargs = generator_run._model_kwargs or {}
    if after_gen:
        model_kwargs = _combine_model_kwargs_and_state(
            generator_run=generator_run, model_class=model.model_class
        )
    bridge_kwargs = generator_run._bridge_kwargs or {}
    model_kwargs = _decode_callables_from_references(model_kwargs)
    bridge_kwargs = _decode_callables_from_references(bridge_kwargs)
    model_keywords = list(model_kwargs.keys())
    for key in model_keywords:
        if key in bridge_kwargs:
            logger.debug(
                f"Keyword argument `{key}` occurs in both model and model bridge "
                f"kwargs stored in the generator run. Assuming the `{key}` kwarg "
                "is passed into the model by the model bridge and removing its "
                "value from the model kwargs."
            )
            del model_kwargs[key]
    return model(experiment=experiment, data=data, **bridge_kwargs, **model_kwargs)


def _combine_model_kwargs_and_state(
    generator_run: GeneratorRun,
    model_class: Type[Model],
    model_kwargs: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """Produces a combined dict of model kwargs and model state after gen,
    extracted from generator run. If model kwargs are not specified,
    model kwargs from the generator run will be used.
    """
    model_kwargs = model_kwargs or generator_run._model_kwargs or {}
    if generator_run._model_state_after_gen is None:
        return model_kwargs

    # We don't want to update `model_kwargs` on the `GenerationStep`,
    # just to add to them for the purpose of this function.
    return {
        **model_kwargs,
        **_extract_model_state_after_gen(
            generator_run=generator_run, model_class=model_class
        ),
    }


def _extract_model_state_after_gen(
    generator_run: GeneratorRun, model_class: Type[Model]
) -> Dict[str, Any]:
    """Extracts serialized post-generation model state from a generator run and
    deserializes it. Fails if no post-generation model state was specified on the
    generator run.
    """
    serialized_model_state = not_none(generator_run._model_state_after_gen)
    # We don't want to update `model_kwargs` on the `GenerationStep`,
    # just to add to them for the purpose of this function.
    return model_class.deserialize_state(serialized_model_state)


def _encode_callables_as_references(kwarg_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Converts callables to references of form <module>.<qualname>, and returns
    the resulting dictionary.
    """
    return {
        k: (
            {"is_callable_as_path": True, "value": callable_to_reference(v)}
            if isfunction(v)
            else v
        )
        for k, v in kwarg_dict.items()
    }


def _decode_callables_from_references(kwarg_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Retrieves callables from references of form <module>.<qualname>, and returns
    the resulting dictionary.
    """
    return {
        k: (
            callable_from_reference(checked_cast(str, v.get("value")))
            if isinstance(v, dict) and v.get("is_callable_as_path", False)
            else v
        )
        for k, v in kwarg_dict.items()
    }



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import importlib
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Type, Union

import torch
from ax.benchmark.runners.base import BenchmarkRunner
from ax.core.arm import Arm
from ax.core.base_trial import BaseTrial, TrialStatus
from ax.utils.common.base import Base
from ax.utils.common.equality import equality_typechecker
from ax.utils.common.serialization import TClassDecoderRegistry, TDecoderRegistry
from ax.utils.common.typeutils import checked_cast
from botorch.test_functions.base import BaseTestProblem, ConstrainedBaseTestProblem
from botorch.test_functions.multi_objective import MultiObjectiveTestProblem
from botorch.utils.transforms import normalize, unnormalize
from torch import Tensor


class BotorchTestProblemRunner(BenchmarkRunner):
    """A Runner for evaluating Botorch BaseTestProblems.

    Given a trial the Runner will evaluate the BaseTestProblem.forward method for each
    arm in the trial, as well as return some metadata about the underlying Botorch
    problem such as the noise_std. We compute the full result on the Runner (as opposed
    to the Metric as is typical in synthetic test problems) because the BoTorch problem
    computes all metrics in one stacked tensor in the MOO case, and we wish to avoid
    recomputation per metric.
    """

    test_problem: BaseTestProblem
    _is_constrained: bool
    _test_problem_class: Type[BaseTestProblem]
    _test_problem_kwargs: Optional[Dict[str, Any]]

    def __init__(
        self,
        test_problem_class: Type[BaseTestProblem],
        test_problem_kwargs: Dict[str, Any],
        outcome_names: List[str],
        modified_bounds: Optional[List[Tuple[float, float]]] = None,
    ) -> None:
        """Initialize the test problem runner.

        Args:
            test_problem_class: The BoTorch test problem class.
            test_problem_kwargs: The keyword arguments used for initializing the
                test problem.
            outcome_names: The names of the outcomes returned by the problem.
            modified_bounds: The bounds that are used by the Ax search space
                while optimizing the problem. If different from the bounds of the
                test problem, we project the parameters into the test problem
                bounds before evaluating the test problem.
                For example, if the test problem is defined on [0, 1] but the Ax
                search space is integers in [0, 10], an Ax parameter value of
                5 will correspond to 0.5 while evaluating the test problem.
                If modified bounds are not provided, the test problem will be
                evaluated using the raw parameter values.
        """

        self._test_problem_class = test_problem_class
        self._test_problem_kwargs = test_problem_kwargs

        # pyre-fixme [45]: Invalid class instantiation
        self.test_problem = test_problem_class(**test_problem_kwargs).to(
            dtype=torch.double
        )
        self._is_constrained: bool = isinstance(
            self.test_problem, ConstrainedBaseTestProblem
        )
        self._is_moo: bool = isinstance(self.test_problem, MultiObjectiveTestProblem)
        self._outcome_names = outcome_names
        self._modified_bounds = modified_bounds

    @property
    def outcome_names(self) -> List[str]:
        return self._outcome_names

    @equality_typechecker
    def __eq__(self, other: Base) -> bool:
        if not isinstance(other, BotorchTestProblemRunner):
            return False

        return (
            self.test_problem.__class__.__name__
            == other.test_problem.__class__.__name__
        )

    def get_noise_stds(self) -> Union[None, float, Dict[str, float]]:
        noise_std = self.test_problem.noise_std
        noise_std_dict: Dict[str, float] = {}
        num_obj = 1 if not self._is_moo else self.test_problem.num_objectives

        # populate any noise_stds for constraints
        if self._is_constrained:
            constraint_noise_std = self.test_problem.constraint_noise_std
            if isinstance(constraint_noise_std, list):
                for i, cns in enumerate(constraint_noise_std, start=num_obj):
                    if cns is not None:
                        noise_std_dict[self.outcome_names[i]] = cns
            elif constraint_noise_std is not None:
                noise_std_dict[self.outcome_names[num_obj]] = constraint_noise_std

        # if none of the constraints are subject to noise, then we may return
        # a single float or None for the noise level

        if not noise_std_dict and not isinstance(noise_std, list):
            return noise_std  # either a float or None

        if isinstance(noise_std, list):
            if not len(noise_std) == num_obj:
                # this shouldn't be possible due to validation upon construction
                # of the multi-objective problem, but better safe than sorry
                raise ValueError(
                    "Noise std must have length equal to number of objectives."
                )
        else:
            noise_std = [noise_std for _ in range(num_obj)]

        for i, noise_std_ in enumerate(noise_std):
            if noise_std_ is not None:
                noise_std_dict[self.outcome_names[i]] = noise_std_

        return noise_std_dict

    def get_Y_true(self, arm: Arm) -> Tensor:
        """Converts X to original bounds -- only if modified bounds were provided --
        and evaluates the test problem. See `__init__` docstring for details.

        Args:
            X: A `batch_shape x d`-dim tensor of point(s) at which to evaluate the
                test problem.

        Returns:
            A `batch_shape x m`-dim tensor of ground truth (noiseless) evaluations.
        """
        X = torch.tensor(
            [
                value
                for _key, value in [*arm.parameters.items()][: self.test_problem.dim]
            ],
            dtype=torch.double,
        )

        if self._modified_bounds is not None:
            # Normalize from modified bounds to unit cube.
            unit_X = normalize(
                X, torch.tensor(self._modified_bounds, dtype=torch.double).T
            )
            # Unnormalize from unit cube to original problem bounds.
            X = unnormalize(unit_X, self.test_problem.bounds)

        Y_true = self.test_problem.evaluate_true(X).view(-1)
        # `BaseTestProblem.evaluate_true()` does not negate the outcome
        if self.test_problem.negate:
            Y_true = -Y_true

        if self._is_constrained:
            # Convention: Concatenate objective and black box constraints. `view()`
            # makes the inputs 1d, so the resulting `Y_true` are also 1d.
            Y_true = torch.cat(
                [Y_true, self.test_problem.evaluate_slack_true(X).view(-1)],
                dim=-1,
            )

        return Y_true

    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> Dict[TrialStatus, Set[int]]:
        return {TrialStatus.COMPLETED: {t.index for t in trials}}

    @classmethod
    # pyre-fixme [2]: Parameter `obj` must have a type other than `Any``
    def serialize_init_args(cls, obj: Any) -> Dict[str, Any]:
        """Serialize the properties needed to initialize the runner.
        Used for storage.
        """
        runner = checked_cast(BotorchTestProblemRunner, obj)

        return {
            "test_problem_module": runner._test_problem_class.__module__,
            "test_problem_class_name": runner._test_problem_class.__name__,
            "test_problem_kwargs": runner._test_problem_kwargs,
            "outcome_names": runner._outcome_names,
            "modified_bounds": runner._modified_bounds,
        }

    @classmethod
    def deserialize_init_args(
        cls,
        args: Dict[str, Any],
        decoder_registry: Optional[TDecoderRegistry] = None,
        class_decoder_registry: Optional[TClassDecoderRegistry] = None,
    ) -> Dict[str, Any]:
        """Given a dictionary, deserialize the properties needed to initialize the
        runner. Used for storage.
        """

        module = importlib.import_module(args["test_problem_module"])

        return {
            "test_problem_class": getattr(module, args["test_problem_class_name"]),
            "test_problem_kwargs": args["test_problem_kwargs"],
            "outcome_names": args["outcome_names"],
            "modified_bounds": args["modified_bounds"],
        }



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from typing import List, Sized

import numpy as np
from ax.core.observation import ObservationFeatures
from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.parameter_constraint import ParameterConstraint
from ax.core.search_space import RobustSearchSpace, SearchSpace
from ax.modelbridge.transforms.choice_encode import (
    ChoiceToNumericChoice,
    OrderedChoiceEncode,
    OrderedChoiceToIntegerRange,
)
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast
from ax.utils.testing.core_stubs import get_robust_search_space


class ChoiceEncodeTransformTest(TestCase):
    t_class = ChoiceToNumericChoice

    def setUp(self) -> None:
        super().setUp()
        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "x", lower=1, upper=3, parameter_type=ParameterType.FLOAT
                ),
                RangeParameter("a", lower=1, upper=2, parameter_type=ParameterType.INT),
                ChoiceParameter(
                    "b",
                    parameter_type=ParameterType.FLOAT,
                    values=[1.0, 10.0, 100.0],
                    is_ordered=True,
                ),
                ChoiceParameter(
                    "c",
                    parameter_type=ParameterType.FLOAT,
                    values=[10.0, 100.0, 1000.0],
                    is_ordered=True,
                    sort_values=False,
                ),
                ChoiceParameter(
                    "d",
                    parameter_type=ParameterType.STRING,
                    values=["r", "q", "z"],
                    sort_values=True,
                ),
                ChoiceParameter(
                    "e",
                    parameter_type=ParameterType.STRING,
                    values=["r", "q", "z"],
                    is_ordered=False,
                    sort_values=False,
                ),
            ],
            parameter_constraints=[
                ParameterConstraint(constraint_dict={"x": -0.5, "a": 1}, bound=0.5)
            ],
        )
        self.t = self.t_class(
            search_space=self.search_space,
            observations=[],
        )
        self.observation_features = [
            ObservationFeatures(
                parameters={"x": 2.2, "a": 2, "b": 10.0, "c": 10.0, "d": "r"}
            )
        ]
        # expected parameters after transform
        self.expected_transformed_params = {
            "x": 2.2,
            "a": 2,
            # ordered float choice originally; transformed normalized value
            "b": normalize_values([1.0, 10.0, 100.0])[1],
            # ordered float choice originally; transformed normalized value
            "c": normalize_values([10.0, 100.0, 1000.0])[0],
            # string choice originally; transformed to int index.
            "d": 1,
        }

    def test_Init(self) -> None:
        self.assertEqual(list(self.t.encoded_parameters.keys()), ["b", "c", "d", "e"])

    def test_TransformObservationFeatures(self) -> None:
        observation_features = self.observation_features
        obs_ft2 = deepcopy(observation_features)
        obs_ft2 = self.t.transform_observation_features(obs_ft2)
        self.assertEqual(
            obs_ft2,
            [ObservationFeatures(parameters=self.expected_transformed_params)],
        )
        obs_ft2 = self.t.untransform_observation_features(obs_ft2)
        self.assertEqual(obs_ft2, observation_features)
        # Test transform on partial features
        obs_ft3 = [ObservationFeatures(parameters={"x": 2.2, "b": 10.0})]
        obs_ft3 = self.t.transform_observation_features(obs_ft3)
        self.assertEqual(
            obs_ft3[0],
            ObservationFeatures(
                parameters={"x": 2.2, "b": self.expected_transformed_params["b"]}
            ),
        )
        obs_ft5 = self.t.transform_observation_features([ObservationFeatures({})])
        self.assertEqual(obs_ft5[0], ObservationFeatures({}))

    def test_ItPreservesChoiceParameterArgs(self) -> None:
        ss2 = deepcopy(self.search_space)
        ss2 = self.t.transform_search_space(ss2)
        for p in ("d", "e"):
            with self.subTest(p):
                tranformed_param = checked_cast(ChoiceParameter, ss2.parameters[p])
                original_param = checked_cast(
                    ChoiceParameter, self.search_space.parameters[p]
                )
                self.assertEqual(tranformed_param.is_ordered, original_param.is_ordered)
                self.assertEqual(
                    tranformed_param.sort_values,
                    original_param.sort_values,
                )
                if self.t_class == ChoiceToNumericChoice:
                    self.assertEqual(
                        tranformed_param.values,
                        [i for i, _ in enumerate(original_param.values)],
                    )
                else:
                    self.assertEqual(
                        tranformed_param.values,
                        original_param.values,
                    )

    def test_TransformSearchSpace(self) -> None:
        ss2 = deepcopy(self.search_space)
        ss2 = self.t.transform_search_space(ss2)

        for p in ("x", "a"):
            self.assertIsInstance(ss2.parameters[p], RangeParameter)
        for p in ("b", "c", "d"):
            self.assertIsInstance(ss2.parameters[p], ChoiceParameter)
        for p in ("x", "b", "c"):
            self.assertEqual(ss2.parameters[p].parameter_type, ParameterType.FLOAT)
        for p in ("a", "d"):
            self.assertEqual(ss2.parameters[p].parameter_type, ParameterType.INT)

        self.assertEqual(
            ss2.parameters["b"].values, normalize_values([1.0, 10.0, 100.0])
        )
        self.assertEqual(
            ss2.parameters["c"].values, normalize_values([10.0, 100.0, 1000.0])
        )
        self.assertEqual(ss2.parameters["d"].values, [0, 1, 2])

        # Ensure we error if we try to transform a fidelity parameter
        ss3 = SearchSpace(
            parameters=[
                ChoiceParameter(
                    "b",
                    parameter_type=ParameterType.FLOAT,
                    values=[1.0, 10.0, 100.0],
                    is_ordered=True,
                    is_fidelity=True,
                    target_value=100.0,
                )
            ]
        )
        t = OrderedChoiceToIntegerRange(search_space=ss3, observations=[])
        with self.assertRaises(ValueError):
            t.transform_search_space(ss3)

    def test_w_parameter_distributions(self) -> None:
        rss = get_robust_search_space()
        # pyre-fixme[16]: `Parameter` has no attribute `_is_ordered`.
        rss.parameters["c"]._is_ordered = True
        # Transform a non-distributional parameter.
        t = self.t_class(
            search_space=rss,
            observations=[],
        )
        rss_new = t.transform_search_space(rss)
        # Make sure that the return value is still a RobustSearchSpace.
        self.assertIsInstance(rss_new, RobustSearchSpace)
        self.assertEqual(set(rss.parameters.keys()), set(rss_new.parameters.keys()))
        # pyre-fixme[16]: `SearchSpace` has no attribute `parameter_distributions`.
        self.assertEqual(rss.parameter_distributions, rss_new.parameter_distributions)
        # pyre-fixme[16]: Optional type has no attribute `parameter_type`.
        self.assertEqual(rss_new.parameters.get("c").parameter_type, ParameterType.INT)
        # Test with environmental variables.
        all_params = list(rss.parameters.values())
        rss = RobustSearchSpace(
            parameters=all_params[2:],
            parameter_distributions=rss.parameter_distributions,
            num_samples=rss.num_samples,
            environmental_variables=all_params[:2],
        )
        t = self.t_class(
            search_space=rss,
            observations=[],
        )
        rss_new = t.transform_search_space(rss)
        self.assertIsInstance(rss_new, RobustSearchSpace)
        self.assertEqual(set(rss.parameters.keys()), set(rss_new.parameters.keys()))
        self.assertEqual(rss.parameter_distributions, rss_new.parameter_distributions)
        # pyre-fixme[16]: `SearchSpace` has no attribute `_environmental_variables`.
        self.assertEqual(rss._environmental_variables, rss_new._environmental_variables)
        self.assertEqual(rss_new.parameters.get("c").parameter_type, ParameterType.INT)


class OrderedChoiceToIntegerRangeTransformTest(ChoiceEncodeTransformTest):
    t_class = OrderedChoiceToIntegerRange

    def setUp(self) -> None:
        super().setUp()
        # expected parameters after transform
        self.expected_transformed_params = {
            "x": 2.2,
            "a": 2,
            # float choice originally; transformed to int index.
            "b": 1,
            # float choice originally; transformed to int index.
            "c": 0,
            "d": "r",
        }

    def test_Init(self) -> None:
        self.assertEqual(list(self.t.encoded_parameters.keys()), ["b", "c"])

    def test_TransformSearchSpace(self) -> None:
        ss2 = deepcopy(self.search_space)
        ss2 = self.t.transform_search_space(ss2)

        for p in ("x", "a", "b", "c"):
            self.assertIsInstance(ss2.parameters[p], RangeParameter)
        self.assertIsInstance(ss2.parameters["d"], ChoiceParameter)
        self.assertEqual(ss2.parameters["x"].parameter_type, ParameterType.FLOAT)
        for p in ("a", "b", "c"):
            self.assertEqual(ss2.parameters[p].parameter_type, ParameterType.INT)
        self.assertEqual(ss2.parameters["d"].parameter_type, ParameterType.STRING)

        self.assertEqual(ss2.parameters["b"].lower, 0)
        self.assertEqual(ss2.parameters["b"].upper, 2)
        self.assertEqual(ss2.parameters["c"].lower, 0)
        self.assertEqual(ss2.parameters["c"].upper, 2)
        self.assertEqual(ss2.parameters["d"].values, ["q", "r", "z"])

        # Ensure we error if we try to transform a fidelity parameter
        ss3 = SearchSpace(
            parameters=[
                ChoiceParameter(
                    "b",
                    parameter_type=ParameterType.FLOAT,
                    values=[1.0, 10.0, 100.0],
                    is_ordered=True,
                    is_fidelity=True,
                    target_value=100.0,
                )
            ]
        )
        t = OrderedChoiceToIntegerRange(search_space=ss3, observations=[])
        with self.assertRaises(ValueError):
            t.transform_search_space(ss3)

    def test_deprecated_OrderedChoiceEncode(self) -> None:
        # Ensure we error if we try to transform a fidelity parameter
        ss3 = SearchSpace(
            parameters=[
                ChoiceParameter(
                    "b",
                    parameter_type=ParameterType.FLOAT,
                    values=[1.0, 10.0, 100.0],
                    is_ordered=True,
                    is_fidelity=True,
                    target_value=100.0,
                )
            ]
        )
        t = OrderedChoiceToIntegerRange(search_space=ss3, observations=[])
        t_deprecated = OrderedChoiceEncode(search_space=ss3, observations=[])
        self.assertEqual(t.__dict__, t_deprecated.__dict__)


def normalize_values(values: Sized) -> List[float]:
    values = np.array(values, dtype=float)
    vmin, vmax = values.min(), values.max()
    if len(values) > 1:
        values = (values - vmin) / (vmax - vmin)
    return values.tolist()
